from functools import reduce
from operator import add
from typing import List, Optional, Type, Union

import networkx as nx
from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.core.models import Path, SubQuery, Table
from sqllineage.core.parser.sqlfluff.models import SqlFluffSubQuery, SqlFluffTable
from sqllineage.core.parser.sqlfluff.utils import (
    find_from_expression_element,
    find_table_identifier,
    is_subquery,
    list_child_segments,
    list_join_clause,
    list_subqueries,
)
from sqllineage.utils.constant import NodeTag
from sqllineage.utils.entities import AnalyzerContext, SubQueryTuple
from sqllineage.utils.helpers import escape_identifier_name


class BaseExtractor:
    """
    Abstract class implementation for extract 'SubQueryLineageHolder' from different statement types
    """

    SUPPORTED_STMT_TYPES: List[str] = []

    def __init__(self, dialect: str, metadata_provider: MetaDataProvider):
        self.dialect = dialect
        self.metadata_provider = metadata_provider

    def can_extract(self, statement_type: str) -> bool:
        """
        Determine if the current lineage holder extractor can process the statement
        :param statement_type: a sqlfluff segment type
        """
        return statement_type in self.SUPPORTED_STMT_TYPES

    def extract(
        self,
        statement: BaseSegment,
        context: AnalyzerContext,
    ) -> SubQueryLineageHolder:
        """
        Extract lineage for a given statement.
        :param statement: a sqlfluff segment with a statement
        :param context: 'AnalyzerContext'
        :return 'SubQueryLineageHolder' object
        """
        raise NotImplementedError

    @classmethod
    def find_table(cls, segment: BaseSegment) -> Optional[Table]:
        table = None
        if segment.type in ["table_reference", "object_reference"]:
            table = SqlFluffTable.of(segment)
        return table

    @classmethod
    def list_subquery(cls, segment: BaseSegment) -> List[SubQuery]:
        """
        The parse_subquery function takes a segment as an argument.
        :param segment: segment to determine if it is a subquery
        :return: A list of `SqlFluffSubQuery` objects, otherwise, if the segment is not matching any of the expected
        types it returns an empty list.
        """
        result: List[SubQuery] = []
        identifiers = segment.get_children("from_expression")
        if identifiers and len(identifiers) > 1:
            # for SQL89 style of JOIN or multiple CTEs, this is actually SubQueries
            return reduce(
                add,
                [
                    cls._parse_subquery(list_subqueries(identifier))
                    for identifier in identifiers
                ],
                [],
            )
        if segment.type in ["select_clause", "from_clause", "where_clause"]:
            result = cls._parse_subquery(list_subqueries(segment))
        elif is_subquery(segment):
            # Parenthesis for SubQuery without alias, this is valid syntax for certain SQL dialect
            result = [SqlFluffSubQuery.of(segment, None)]
        return result

    def _list_table_from_from_clause_or_join_clause(
        self, segment: BaseSegment, holder: SubQueryLineageHolder
    ) -> List[Union[Table, SubQuery, Path]]:
        """
        Extract table from from_clause or join_clause, join_clause is a child node of from_clause.
        """
        tables = []
        if segment.type in ["from_clause", "join_clause", "update_statement"]:
            if len(from_expressions := segment.get_children("from_expression")) > 1:
                # SQL89 style of join
                for from_expression in from_expressions:
                    if from_expression_element := find_from_expression_element(
                        from_expression
                    ):
                        tables += self._add_dataset_from_expression_element(
                            from_expression_element, holder
                        )
            else:
                if from_expression_element := find_from_expression_element(segment):
                    tables += self._add_dataset_from_expression_element(
                        from_expression_element, holder
                    )
                for join_clause in list_join_clause(segment):
                    tables += self._list_table_from_from_clause_or_join_clause(
                        join_clause, holder
                    )

        return tables

    @staticmethod
    def _add_dataset_from_expression_element(
        segment: BaseSegment, holder: SubQueryLineageHolder
    ) -> List[Union[Table, SubQuery, Path]]:
        """
        Append tables and subqueries identified in the 'from_expression_element' type segment to the table and
        holder extra subqueries sets
        """
        tables: List[Union[Table, SubQuery, Path]] = []
        all_segments = [
            seg for seg in list_child_segments(segment) if seg.type != "keyword"
        ]
        if table_expression := segment.get_child("table_expression"):
            if table_expression.get_child("function"):
                # for UNNEST or generator function, no dataset involved
                return tables
        first_segment = all_segments[0]
        if first_segment.type == "bracketed":
            if table_expression := first_segment.get_child("table_expression"):
                if table_expression.get_child("values_clause"):
                    # (VALUES ...) AS alias, no dataset involved
                    return tables
        subqueries = list_subqueries(segment)
        if subqueries:
            for sq in subqueries:
                bracketed, alias = sq
                read_sq = SqlFluffSubQuery.of(bracketed, alias)
                tables.append(read_sq)
        else:
            table_identifier = find_table_identifier(segment)
            if table_identifier:
                subquery_flag = False
                alias = None
                if len(all_segments) > 1 and all_segments[1].type == "alias_expression":
                    all_segments = list_child_segments(all_segments[1])
                    alias = str(
                        all_segments[1].raw
                        if len(all_segments) > 1
                        else all_segments[0].raw
                    )
                if "." not in table_identifier.raw:
                    cte_dict = {s.alias: s for s in holder.cte}
                    cte = cte_dict.get(escape_identifier_name(table_identifier.raw))
                    if cte is not None:
                        # could reference CTE with or without alias
                        tables.append(
                            SqlFluffSubQuery.of(
                                cte.query,
                                alias or table_identifier.raw,
                            )
                        )
                        subquery_flag = True
                if subquery_flag is False:
                    if table_identifier.type == "file_reference":
                        tables.append(
                            Path(
                                escape_identifier_name(
                                    table_identifier.segments[-1].raw
                                )
                            )
                        )
                    else:
                        tables.append(SqlFluffTable.of(table_identifier, alias=alias))

        return tables

    @classmethod
    def _parse_subquery(cls, subqueries: List[SubQueryTuple]) -> List[SubQuery]:
        """
        Convert a list of 'SqlFluffSubQueryTuple' to 'SqlFluffSubQuery'
        :param subqueries:  a list of 'SqlFluffSubQueryTuple'
        :return:  a list of 'SqlFluffSubQuery'
        """
        return [
            SqlFluffSubQuery.of(bracketed_segment, alias)
            for bracketed_segment, alias in subqueries
        ]

    def delegate_to(
        self,
        extractor_cls: "Type[BaseExtractor]",
        segment: BaseSegment,
        context: AnalyzerContext,
    ) -> SubQueryLineageHolder:
        """
        delegate to another type of extractor to extract
        """
        return extractor_cls(self.dialect, self.metadata_provider).extract(
            segment, context
        )

    def extract_subquery(
        self, subqueries: List[SubQuery], holder: SubQueryLineageHolder
    ):
        """
        extract subqueries collected from statement-level segment
        """
        from .cte import CteExtractor
        from .select import SelectExtractor

        for sq in subqueries:
            extractor_cls = (
                CteExtractor
                if sq.query.get_child("with_compound_statement")
                else SelectExtractor
            )
            subquery_holder = extractor_cls(
                self.dialect, self.metadata_provider
            ).extract(sq.query, AnalyzerContext(cte=holder.cte, write={sq}))
            # remove WRITE tag from subquery so that the combined holder won't have multiple WRITE dataset
            nx.set_node_attributes(subquery_holder.graph, {sq: False}, NodeTag.WRITE)
            holder |= subquery_holder

    @staticmethod
    def _init_holder(context: AnalyzerContext) -> SubQueryLineageHolder:
        """
        Initialize lineage holder for a given 'AnalyzerContext'
        :param context: a previous context that the lineage extractor must consider
        :return: an initialized SubQueryLineageHolder
        """
        holder = SubQueryLineageHolder()

        if context.cte is not None:
            # CTE can be referenced by subsequent CTEs
            for cte in context.cte:
                holder.add_cte(cte)

        if context.write is not None:
            # If within subquery, then manually add subquery as target table
            for write in context.write:
                holder.add_write(write)

        if context.write_columns:
            # target columns can be referred while creating column level lineage
            holder.add_write_column(*context.write_columns)

        return holder



import re
from typing import Optional, Union

from sqlparse.sql import (
    Case,
    Function,
    Identifier,
    IdentifierList,
    Operation,
    Parenthesis,
    Token,
)
from sqlparse.tokens import Literal, Wildcard

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.models import Path, SubQuery, Table
from sqllineage.core.parser import SourceHandlerMixin
from sqllineage.core.parser.sqlparse.handlers.base import NextTokenBaseHandler
from sqllineage.core.parser.sqlparse.models import (
    SqlParseColumn,
    SqlParseSubQuery,
    SqlParseTable,
)
from sqllineage.core.parser.sqlparse.utils import (
    get_subquery_parentheses,
    is_subquery,
    is_values_clause,
)
from sqllineage.exceptions import SQLLineageException
from sqllineage.utils.helpers import escape_identifier_name


class SourceHandler(SourceHandlerMixin, NextTokenBaseHandler):
    """Source Table & Column Handler."""

    SOURCE_TABLE_TOKENS = (
        r"FROM",
        # inspired by https://github.com/andialbrecht/sqlparse/blob/master/sqlparse/keywords.py
        r"((LEFT\s+|RIGHT\s+|FULL\s+)?(INNER\s+|OUTER\s+|STRAIGHT\s+)?|(CROSS\s+|NATURAL\s+)?)?JOIN",
    )

    def __init__(self):
        self.column_flag = False
        self.columns = []
        self.tables = []
        self.union_barriers = []
        super().__init__()

    def _indicate(self, token: Token) -> bool:
        if token.normalized in ("UNION", "UNION ALL"):
            self.union_barriers.append((len(self.columns), len(self.tables)))

        if self.column_flag is True and bool(token.normalized == "DISTINCT"):
            # special handling for SELECT DISTINCT
            return True

        if any(re.match(regex, token.normalized) for regex in self.SOURCE_TABLE_TOKENS):
            self.column_flag = False
            return True
        elif bool(token.normalized == "SELECT"):
            self.column_flag = True
            return True
        else:
            return False

    def _handle(self, token: Token, holder: SubQueryLineageHolder) -> None:
        if self.column_flag:
            self._handle_column(token)
        else:
            self._handle_table(token, holder)

    def _handle_table(self, token: Token, holder: SubQueryLineageHolder) -> None:
        if isinstance(token, Identifier):
            self._add_dataset_from_identifier(token, holder)
        elif isinstance(token, IdentifierList):
            # This is to support join in ANSI-89 syntax
            for identifier in token.get_sublists():
                self._add_dataset_from_identifier(identifier, holder)
        elif isinstance(token, Parenthesis):
            if is_subquery(token):
                # SELECT col1 FROM (SELECT col2 FROM tab1), the subquery will be parsed as Parenthesis
                # This syntax without alias for subquery is invalid in MySQL, while valid for SparkSQL
                self.tables.append(SqlParseSubQuery.of(token, None))
            elif is_values_clause(token):
                # SELECT * FROM (VALUES ...), no operation needed
                pass
            else:
                # SELECT * FROM (tab2) or SELECT * FROM (tab2 JOIN tab1), which is valid syntax
                using_flag = False
                for t in token.tokens:
                    if t.is_keyword and t.normalized == "USING":
                        using_flag = True
                    if isinstance(t, Identifier) or (
                        isinstance(t, Parenthesis) and using_flag is False
                    ):
                        self._handle(t, holder)
        elif token.ttype == Literal.String.Single:
            self.tables.append(Path(token.value))
        elif isinstance(token, Function):
            # functions like unnest or generator can output a sequence of values as source, ignore it for now
            pass
        else:
            raise SQLLineageException(
                "An Identifier is expected, got %s[value: %s] instead."
                % (type(token).__name__, token)
            )

    def _handle_column(self, token: Token) -> None:
        column_token_types = (Identifier, Function, Operation, Case, Parenthesis)
        if isinstance(token, column_token_types) or token.ttype is Wildcard:
            column_tokens = [token]
        elif isinstance(token, IdentifierList):
            column_tokens = [
                sub_token
                for sub_token in token.tokens
                if (
                    isinstance(sub_token, column_token_types)
                    and not sub_token.value.startswith(
                        "@"
                    )  # ignore tsql variable column starts with @
                )
                or sub_token.ttype is Wildcard  # For case like: select a.col, b.*
            ]
        else:
            # SELECT constant value will end up here
            column_tokens = []
        for token in column_tokens:
            self.columns.append(SqlParseColumn.of(token))

    def _add_dataset_from_identifier(
        self, identifier: Identifier, holder: SubQueryLineageHolder
    ) -> None:
        first_token = identifier.token_first(skip_cm=True)
        if isinstance(first_token, Function):
            # function() as alias, no dataset involved
            return None
        elif isinstance(first_token, Parenthesis) and is_values_clause(first_token):
            # (VALUES ...) AS alias, no dataset involved
            return None
        dataset: Union[Table, SubQuery, Path]
        path_match = re.match(r"(parquet|csv|json)\.`(.*)`", identifier.value)
        if path_match is not None:
            dataset = Path(path_match.groups()[1])
        else:
            read: Optional[Union[Table, SubQuery]] = None
            subqueries = get_subquery_parentheses(identifier)
            if len(subqueries) > 0:
                # SELECT col1 FROM (SELECT col2 FROM tab1) dt, the subquery will be parsed as Identifier
                # referring https://github.com/andialbrecht/sqlparse/issues/218 for further information
                parenthesis, alias = subqueries[0]
                read = SqlParseSubQuery.of(parenthesis, alias)
            else:
                cte_dict = {s.alias: s for s in holder.cte}
                if "." not in identifier.value:
                    cte = cte_dict.get(
                        escape_identifier_name(identifier.get_real_name())
                    )
                    if cte is not None:
                        # could reference CTE with or without alias
                        read = SqlParseSubQuery.of(
                            cte.query,
                            identifier.get_alias() or identifier.get_real_name(),
                        )
                if read is None:
                    read = SqlParseTable.of(identifier)
            dataset = read
        if dataset:
            self.tables.append(dataset)



from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import StatementLineageHolder
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.utils.entities import AnalyzerContext


class NoopExtractor(BaseExtractor):
    """
    Extractor for queries which do not provide any lineage
    """

    SUPPORTED_STMT_TYPES = [
        "delete_statement",
        "truncate_table",
        "refresh_statement",
        "cache_table",
        "uncache_table",
        "show_statement",
        "describe_statement",
        "use_statement",
        "declare_segment",
        "analyze_statement",
        "add_jar_statement",
        "create_function_statement",
        "drop_function_statement",
        "set_statement",
    ]

    def extract(
        self,
        statement: BaseSegment,
        context: AnalyzerContext,
    ) -> StatementLineageHolder:
        return StatementLineageHolder()



from sqlparse.sql import Function, Identifier, IdentifierList, Token

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.parser.sqlparse.handlers.base import NextTokenBaseHandler
from sqllineage.core.parser.sqlparse.models import SqlParseSubQuery


class CTEHandler(NextTokenBaseHandler):
    """Common Table Expression (With Queries) Handler."""

    CTE_TOKENS = ("WITH",)

    def _indicate(self, token: Token) -> bool:
        return token.normalized in self.CTE_TOKENS

    def _handle(self, token: Token, holder: SubQueryLineageHolder) -> None:
        # when CTE used without AS, it will be parsed as Function. This syntax is valid in SparkSQL
        cte_token_types = (Identifier, Function)
        if isinstance(token, cte_token_types):
            cte = [token]
        elif isinstance(token, IdentifierList):
            cte = [
                token for token in token.tokens if isinstance(token, cte_token_types)
            ]
        else:
            # CREATE TABLE tbl1 (col1 VARCHAR) WITH (bucketed_by = ARRAY['col1'], bucket_count = 256).
            # This syntax is valid for bucketing in Trino and not the CTE, token will be Parenthesis here
            cte = []
        for token in cte:
            sublist = list(token.get_sublists())
            if sublist and not (isinstance(sublist[0], Function)):
                # CTE: tbl AS (SELECT 1), tbl is alias and (SELECT 1) is subquery Parenthesis
                holder.add_cte(SqlParseSubQuery.of(sublist[0], token.get_real_name()))



import pytest

from ...helpers import assert_table_lineage_equal


"""
This test class will contain all the tests for testing 'Select Queries' where the dialect is not ANSI.
"""


@pytest.mark.parametrize(
    "dialect", ["athena", "bigquery", "databricks", "hive", "mysql", "sparksql"]
)
def test_select_with_table_name_in_backtick(dialect: str):
    assert_table_lineage_equal("SELECT * FROM `tab1`", {"tab1"}, dialect=dialect)


@pytest.mark.parametrize(
    "dialect", ["athena", "bigquery", "databricks", "hive", "mysql", "sparksql"]
)
def test_select_with_schema_in_backtick(dialect: str):
    assert_table_lineage_equal(
        "SELECT col1 FROM `schema1`.`tab1`", {"schema1.tab1"}, dialect=dialect
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_select_left_semi_join(dialect: str):
    assert_table_lineage_equal(
        "SELECT * FROM tab1 LEFT SEMI JOIN tab2", {"tab1", "tab2"}, dialect=dialect
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_select_left_semi_join_with_on(dialect: str):
    assert_table_lineage_equal(
        "SELECT * FROM tab1 LEFT SEMI JOIN tab2 ON (tab1.col1 = tab2.col2)",
        {"tab1", "tab2"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["snowflake"])
def test_select_from_generator(dialect: str):
    # generator is Snowflake specific
    sql = """SELECT seq4(), uniform(1, 10, random(12))
FROM table(generator()) v
ORDER BY 1;"""
    assert_table_lineage_equal(sql, dialect=dialect)


@pytest.mark.parametrize("dialect", ["postgres", "redshift", "tsql"])
def test_select_into(dialect: str):
    """
    postgres: https://www.postgresql.org/docs/current/sql-selectinto.html
    redshift: https://docs.aws.amazon.com/redshift/latest/dg/r_SELECT_INTO.html
    tsql: https://learn.microsoft.com/en-us/sql/t-sql/queries/select-into-clause-transact-sql?view=sql-server-ver16
    """
    sql = "SELECT * INTO films_recent FROM films WHERE date_prod >= '2002-01-01'"
    assert_table_lineage_equal(sql, {"films"}, {"films_recent"}, dialect=dialect)


@pytest.mark.parametrize("dialect", ["postgres", "tsql"])
def test_select_into_with_union(dialect: str):
    sql = "SELECT * INTO films_all FROM films UNION ALL SELECT * FROM films_backup"
    assert_table_lineage_equal(
        sql, {"films", "films_backup"}, {"films_all"}, dialect=dialect
    )


@pytest.mark.parametrize("dialect", ["athena"])
def test_select_from_unnest_with_ordinality(dialect: str):
    """
    https://prestodb.io/docs/current/sql/select.html#unnest
    """
    sql = """
    SELECT numbers, n, a
    FROM (
      VALUES
        (ARRAY[2, 5]),
        (ARRAY[7, 8, 9])
    ) AS x (numbers)
    CROSS JOIN UNNEST(numbers) WITH ORDINALITY AS t (n, a);
    """
    assert_table_lineage_equal(sql, dialect=dialect)



import os
from unittest.mock import patch

import pytest

from ....helpers import assert_table_lineage_equal


@patch.dict(os.environ, {"SQLLINEAGE_TSQL_NO_SEMICOLON": "TRUE"})
@pytest.mark.parametrize("dialect", ["tsql"])
def test_tsql_multi_statement_no_semicolon(dialect: str):
    """
    tsql multiple statements without explicit semicolon as splitter.
    """
    sql = """insert into tab1 select * from foo
insert into tab2 select * from bar"""
    assert_table_lineage_equal(
        sql,
        {"foo", "bar"},
        {"tab1", "tab2"},
        dialect=dialect,
        test_sqlparse=False,
    )



"""
app is a wsgi application which accepts environ and start_response as argument.
Here We implement a simple flask-like api to avoid explicitly add it as dependency.
wsgiref is used to spawn a server from sqllineage commandline. To serve production traffic, you can/should put
the app behind a real production server like gunicorn or uwsgi, as app is wsgi compatible.
A simple gunicorn example: gunicorn sqllineage.drawing:app
"""

import json
import logging
import mimetypes
import os
from argparse import Namespace
from http import HTTPStatus
from pathlib import Path
from typing import Any, Callable, Dict, List
from urllib.parse import urlencode
from wsgiref.simple_server import make_server

from sqllineage import DEFAULT_DIALECT, DEFAULT_HOST, DEFAULT_PORT, STATIC_FOLDER
from sqllineage.config import SQLLineageConfig
from sqllineage.exceptions import SQLLineageException
from sqllineage.utils.constant import LineageLevel
from sqllineage.utils.helpers import extract_sql_from_args

logger = logging.getLogger(__name__)


class SQLLineageApp:
    def __init__(self) -> None:
        self.routes: Dict[str, Callable[[Dict[str, Any]], Dict[str, Any]]] = {}
        self.root_path = Path(SQLLineageConfig.DIRECTORY)

    def route(self, path: str):
        def wrapper(handler):
            self.routes[path] = handler
            return handler

        return wrapper

    def __call__(self, environ, start_response) -> List[bytes]:
        static_folder = Path(os.path.dirname(__file__)).joinpath(Path(STATIC_FOLDER))
        request_method = environ["REQUEST_METHOD"]
        path_info = environ["PATH_INFO"]
        try:
            if request_method == "GET":
                mimetype = "text/html; charset=utf-8"
                if path_info == "/":
                    static_fname = str(static_folder.joinpath(Path("index.html")))
                else:
                    if ".." in path_info:
                        # Do not allow going back to parent path of static folder
                        return self.handle_404(start_response)
                    static_file = static_folder.joinpath(Path(path_info.strip("/")))
                    if static_file.exists():
                        static_fname = str(static_file)
                        optional_mimetype = mimetypes.guess_type(path_info)[0]
                        mimetype = (
                            optional_mimetype
                            if optional_mimetype is not None
                            else mimetype
                        )
                    else:
                        return self.handle_404(start_response)
                with open(static_fname, "rb") as f:
                    text = f.read()
                return self.handle_200_text(start_response, mimetype, text)
            elif request_method == "POST":
                if path_info in self.routes:
                    request_body_size = int(environ["CONTENT_LENGTH"])
                    request_body = environ["wsgi.input"].read(request_body_size)
                    payload = json.loads(request_body)
                    for param in ["d", "f"]:
                        if param in payload and not str(
                            Path(payload[param]).absolute()
                        ).startswith(str(Path(self.root_path).absolute())):
                            return self.handle_403(start_response)
                    data = self.routes[path_info](payload)
                    return self.handle_200_json(start_response, data)
                else:
                    return self.handle_404(start_response)
            elif request_method == "OPTIONS":
                if path_info in self.routes:
                    start_response(
                        "200 OK",
                        [
                            ("Access-Control-Allow-Origin", "*"),
                            (
                                "Access-Control-Allow-Headers",
                                "Content-Type",
                            ),
                            ("Access-Control-Allow-Methods", "POST"),
                        ],
                    )
                    return []
                else:
                    return self.handle_404(start_response)
            else:
                return self.handle_405(start_response)
        except (SystemExit, IsADirectoryError, FileNotFoundError, PermissionError):
            return self.handle_404(start_response)
        except (SQLLineageException, RuntimeError) as e:
            return self.handle_400(start_response, str(e))

    @staticmethod
    def handle_200_text(start_response, mimetype, text) -> List[bytes]:
        status_code = HTTPStatus.OK
        start_response(
            f"{status_code.value} {status_code.phrase}", [("Content-type", mimetype)]
        )
        return [text]

    def handle_200_json(self, start_response, data) -> List[bytes]:
        return self.handle_json_response(start_response, HTTPStatus.OK, data)

    def handle_400(self, start_response, message) -> List[bytes]:
        return self.handle_client_error_response(
            start_response, HTTPStatus.BAD_REQUEST, message
        )

    def handle_403(self, start_response) -> List[bytes]:
        message = "File Not Allowed For Accessing"
        return self.handle_client_error_response(
            start_response, HTTPStatus.FORBIDDEN, message
        )

    def handle_404(self, start_response) -> List[bytes]:
        message = "File Not Found"
        return self.handle_client_error_response(
            start_response, HTTPStatus.NOT_FOUND, message
        )

    def handle_405(self, start_response) -> List[bytes]:
        message = "Method Not Allowed"
        return self.handle_client_error_response(
            start_response, HTTPStatus.METHOD_NOT_ALLOWED, message
        )

    def handle_client_error_response(
        self, start_response, status_code, message
    ) -> List[bytes]:
        data = {"message": message}
        return self.handle_json_response(start_response, status_code, data)

    @staticmethod
    def handle_json_response(start_response, status_code, data) -> List[bytes]:
        start_response(
            f"{status_code.value} {status_code.phrase}",
            [
                ("Content-type", "application/json"),
                ("Access-Control-Allow-Origin", "*"),
            ],
        )
        return [json.dumps(data).encode("utf-8")]


app = SQLLineageApp()


@app.route("/lineage")
def lineage(payload):
    # this is to avoid circular import
    from sqllineage.runner import LineageRunner

    req_args = Namespace(**payload)
    sql = extract_sql_from_args(req_args)
    dialect = getattr(req_args, "dialect", DEFAULT_DIALECT)
    lr = LineageRunner(sql, dialect=dialect, verbose=True)
    data = {
        "verbose": str(lr),
        "dag": lr.to_cytoscape(),
        "column": lr.to_cytoscape(LineageLevel.COLUMN),
    }
    return data


@app.route("/script")
def script(payload):
    req_args = Namespace(**payload)
    sql = extract_sql_from_args(req_args)
    return {"content": sql}


@app.route("/directory")
def directory(payload):
    if payload.get("f"):
        root = Path(payload["f"]).parent
    elif payload.get("d"):
        root = Path(payload["d"])
    else:
        root = Path(SQLLineageConfig.DIRECTORY)
    data = {
        "id": str(root),
        "name": root.name,
        "is_dir": True,
        "children": [
            {"id": str(p), "name": p.name, "is_dir": p.is_dir()}
            for p in sorted(root.iterdir(), key=lambda _: (not _.is_dir(), _.name))
        ],
    }
    return data


def draw_lineage_graph(**kwargs) -> None:
    host = kwargs.pop("host", DEFAULT_HOST)
    port = kwargs.pop("port", DEFAULT_PORT)
    querystring = urlencode({k: v for k, v in kwargs.items() if v})
    path = f"/?{querystring}" if querystring else "/"
    if "f" in kwargs:
        app.root_path = Path(kwargs["f"]).parent
    with make_server(host, port, app) as httpd:
        print(f" * SQLLineage Running on http://{host}:{port}{path}")
        httpd.serve_forever()



import pytest

from sqllineage.core.models import Path
from ...helpers import assert_table_lineage_equal


@pytest.mark.parametrize("dialect", ["postgres", "redshift"])
def test_copy_from_path(dialect: str):
    """
    https://www.postgresql.org/docs/current/sql-copy.html (Postgres)
    https://docs.aws.amazon.com/es_es/redshift/latest/dg/r_COPY.html (Redshift)
    """
    assert_table_lineage_equal(
        "COPY tab1 FROM 's3://mybucket/mypath'",
        {Path("s3://mybucket/mypath")},
        {"tab1"},
        dialect=dialect,
    )


@pytest.mark.parametrize(
    "dialect, path",
    [
        ("snowflake", "s3://mybucket/mypath"),
        ("tsql", "https://myaccount.blob.core.windows.net/myblobcontainer/folder1/"),
    ],
)
def test_copy_into_path(dialect: str, path: str):
    """
    check following link for syntax reference:
        Snowflake: https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html
        Microsoft T-SQL: https://docs.microsoft.com/en-us/sql/t-sql/statements/copy-into-transact-sql?view=azure-sqldw-latest   # noqa
    """
    assert_table_lineage_equal(
        f"COPY INTO tab1 FROM '{path}'",
        {Path(path)},
        {"tab1"},
        dialect=dialect,
    )


@pytest.mark.parametrize("data_source", ["parquet", "json", "csv"])
@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_select_from_files(data_source: str, dialect: str):
    """
    check following link for syntax reference:
        https://spark.apache.org/docs/latest/sql-data-sources-load-save-functions.html#run-sql-on-files-directly
    """
    assert_table_lineage_equal(
        f"SELECT * FROM {data_source}.`examples/src/main/resources/`",
        {Path("examples/src/main/resources/")},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_insert_overwrite_directory(dialect: str):
    """
    check following link for syntax reference:
        https://spark.apache.org/docs/latest/sql-ref-syntax-dml-insert-overwrite-directory.html
    """
    assert_table_lineage_equal(
        """INSERT OVERWRITE DIRECTORY 'hdfs://path/to/folder'
SELECT * FROM tab1""",
        {"tab1"},
        {Path("hdfs://path/to/folder")},
        dialect=dialect,
    )



from ...helpers import assert_table_lineage_equal


def test_swapping_partitions():
    """
    See https://www.vertica.com/docs/10.0.x/HTML/Content/Authoring/AdministratorsGuide/Partitions/SwappingPartitions.htm
    for language specification
    """
    assert_table_lineage_equal(
        "SELECT swap_partitions_between_tables('staging', 'min-range-value', 'max-range-value', 'target')",
        {"staging"},
        {"target"},
    )



from ...helpers import assert_table_lineage_equal


def test_insert_into_values():
    assert_table_lineage_equal("INSERT INTO tab1 VALUES (1, 2)", set(), {"tab1"})


def test_insert_into_values_with_subquery():
    assert_table_lineage_equal(
        "INSERT INTO tab1 VALUES (1, (SELECT max(id) FROM tab2))",
        {"tab2"},
        {"tab1"},
    )


def test_insert_into_values_with_multiple_subquery():
    assert_table_lineage_equal(
        "INSERT INTO tab1 VALUES ((SELECT max(id) FROM tab2), (SELECT max(id) FROM tab3))",
        {"tab2", "tab3"},
        {"tab1"},
    )


def test_insert_into_values_with_multiple_subquery_in_multiple_row():
    assert_table_lineage_equal(
        "INSERT INTO tab1 VALUES (1, (SELECT max(id) FROM tab2)), (2, (SELECT max(id) FROM tab3))",
        {"tab2", "tab3"},
        {"tab1"},
    )


def test_insert_into_select():
    assert_table_lineage_equal(
        "INSERT INTO tab1 SELECT * FROM tab2;",
        {"tab2"},
        {"tab1"},
    )


def test_non_reserved_keyword_as_target():
    assert_table_lineage_equal(
        "INSERT INTO host SELECT col1, col2 FROM segment",
        {"segment"},
        {"host"},
        test_sqlparse=False,
    )


def test_insert_into_qualified_table_with_parenthesized_query():
    """
    For sqlparse, it will work if:
        1) table in unqualified
    OR  2) query is not surrounded by parenthesis
    With both in the game, it breaks.
    """
    sql = """INSERT INTO default.tab2
    (SELECT *
    FROM tab1)"""
    assert_table_lineage_equal(sql, {"tab1"}, {"default.tab2"}, test_sqlparse=False)



import pytest
from sqlparse.sql import Parenthesis

from sqllineage.core.models import Column, Path, Schema, SubQuery, Table
from sqllineage.exceptions import SQLLineageException


def test_repr_dummy():
    assert repr(Schema())
    assert repr(Table(""))
    assert repr(Table("a.b.c"))
    assert repr(SubQuery(Parenthesis(), Parenthesis().value, ""))
    assert repr(Column("a.b"))
    assert repr(Path(""))
    with pytest.raises(SQLLineageException):
        Table("a.b.c.d")
    with pytest.warns(Warning):
        Table("a.b", Schema("c"))


def test_hash_eq():
    assert Schema("a") == Schema("a")
    assert len({Schema("a"), Schema("a")}) == 1
    assert Table("a") == Table("a")
    assert len({Table("a"), Table("a")}) == 1


def test_of_dummy():
    with pytest.raises(NotImplementedError):
        Column.of("")
    with pytest.raises(NotImplementedError):
        Table.of("")
    with pytest.raises(NotImplementedError):
        SubQuery.of("", None)



from ...helpers import assert_table_lineage_equal


def test_use():
    assert_table_lineage_equal("USE db1")


def test_drop():
    assert_table_lineage_equal("DROP TABLE IF EXISTS tab1")


def test_drop_with_comment():
    assert_table_lineage_equal(
        """--comment
DROP TABLE IF EXISTS tab1"""
    )


def test_drop_view():
    assert_table_lineage_equal("DROP VIEW IF EXISTS view1")


def test_alter_table_rename():
    assert_table_lineage_equal("ALTER TABLE tab1 rename TO tab2")


def test_truncate_table():
    assert_table_lineage_equal("TRUNCATE TABLE tab1")


def test_delete_from_table():
    assert_table_lineage_equal("DELETE FROM table tab1")



from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.models import Column
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.core.parser.sqlfluff.utils import (
    extract_column_qualifier,
    list_child_segments,
)
from sqllineage.utils.entities import AnalyzerContext


class UpdateExtractor(BaseExtractor):
    """
    Update statement lineage extractor
    """

    SUPPORTED_STMT_TYPES = ["update_statement"]

    def extract(
        self, statement: BaseSegment, context: AnalyzerContext
    ) -> SubQueryLineageHolder:
        holder = self._init_holder(context)
        tgt_flag = False
        columns = []
        subqueries = []
        for segment in list_child_segments(statement):
            if segment.type == "from_expression":
                # UPDATE with JOIN, mysql only syntax
                # from_expression is wrapped directly under update_statement and there's no from_clause in this case
                # we need to pass update_statement to the function
                if from_join_tables := self._list_table_from_from_clause_or_join_clause(
                    statement, holder
                ):
                    holder.add_write(from_join_tables[0])
                    for join_table in from_join_tables[1:]:
                        holder.add_read(join_table)

            if segment.type == "keyword" and segment.raw_upper == "UPDATE":
                tgt_flag = True
                continue

            if tgt_flag:
                if write_table := self.find_table(segment):
                    holder.add_write(write_table)
                tgt_flag = False

            if segment.type == "set_clause_list":
                for set_clause in segment.get_children("set_clause"):
                    column_references = set_clause.get_children("column_reference")
                    if len(column_references) == 2:
                        tgt_cqt = extract_column_qualifier(column_references[0])
                        src_cqt = extract_column_qualifier(column_references[1])
                        if tgt_cqt is not None and src_cqt is not None:
                            columns.append(
                                Column(tgt_cqt.column, source_columns=[src_cqt])
                            )

            if segment.type == "from_clause":
                # UPDATE FROM, ansi syntax
                # there can be multiple from items, each may be a table or a subquery
                for sq in self.list_subquery(segment):
                    subqueries.append(sq)
                for read_table in self._list_table_from_from_clause_or_join_clause(
                    segment, holder
                ):
                    holder.add_read(read_table)

        for tgt_col in columns:
            tgt_col.parent = list(holder.write)[0]
            for src_col in tgt_col.to_source_columns(
                holder.get_alias_mapping_from_table_group(list(holder.read))
            ):
                holder.add_column_lineage(src_col, tgt_col)

        self.extract_subquery(subqueries, holder)

        return holder



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_select_column_using_function():
    sql = """INSERT INTO tab1
SELECT max(col1),
       count(*)
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("max(col1)", "tab1"),
            ),
            (
                ColumnQualifierTuple("*", "tab2"),
                ColumnQualifierTuple("count(*)", "tab1"),
            ),
        ],
    )
    sql = """INSERT INTO tab1
SELECT max(col1) AS col2,
       count(*)  AS cnt
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col2", "tab1"),
            ),
            (ColumnQualifierTuple("*", "tab2"), ColumnQualifierTuple("cnt", "tab1")),
        ],
    )


def test_select_column_using_function_with_complex_parameter():
    sql = """INSERT INTO tab1
SELECT if(col1 = 'foo' AND col2 = 'bar', 1, 0) AS flag
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("flag", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("flag", "tab1"),
            ),
        ],
    )


def test_select_column_using_window_function():
    sql = """INSERT INTO tab1
SELECT row_number() over (partition BY col1 ORDER BY col2 DESC) AS rnum
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("rnum", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("rnum", "tab1"),
            ),
        ],
    )


def test_select_column_using_window_function_with_parameters():
    sql = """INSERT INTO tab1
SELECT col0,
       max(col3) over (partition BY col1 ORDER BY col2 DESC) AS rnum,
       col4
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col0", "tab2"),
                ColumnQualifierTuple("col0", "tab1"),
            ),
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("rnum", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("rnum", "tab1"),
            ),
            (
                ColumnQualifierTuple("col3", "tab2"),
                ColumnQualifierTuple("rnum", "tab1"),
            ),
            (
                ColumnQualifierTuple("col4", "tab2"),
                ColumnQualifierTuple("col4", "tab1"),
            ),
        ],
    )


def test_column_with_ctas_and_func():
    sql = """CREATE TABLE tab2 AS
SELECT coalesce(col1, 0)             AS col1,
       if(col1 IS NOT NULL, 1, NULL) AS col2
FROM tab1"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab1"),
                ColumnQualifierTuple("col1", "tab2"),
            ),
            (
                ColumnQualifierTuple("col1", "tab1"),
                ColumnQualifierTuple("col2", "tab2"),
            ),
        ],
    )


def test_coalesce_with_whitespace():
    """
    coalesce is a keyword since ANSI SQL-2023
    usually it's parsed as a function. however, when whitespace followed which is valid syntax,
    sqlparse cannot produce the correct AST
    """
    sql = """INSERT INTO tab1
SELECT coalesce (col1, col2) as col3
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
        test_sqlparse=False,
    )



from sqllineage.utils.helpers import split


def test_split_statements():
    sql = "SELECT * FROM tab1; SELECT * FROM tab2;"
    assert len(split(sql)) == 2


def test_split_statements_with_heading_and_ending_new_line():
    sql = "\nSELECT * FROM tab1;\nSELECT * FROM tab2;\n"
    assert len(split(sql)) == 2


def test_split_statements_with_comment():
    sql = """SELECT 1;

-- SELECT 2;"""
    assert len(split(sql)) == 1


def test_split_statements_with_show_create_table():
    sql = """SELECT 1;

SHOW CREATE TABLE tab1;"""
    assert len(split(sql)) == 2


def test_split_statements_with_desc():
    sql = """SELECT 1;

DESC tab1;"""
    assert len(split(sql)) == 2


def test_split_statement_ends_with_multiple_semicolons():
    sql = "SELECT 1;;;"
    assert len(split(sql)) == 1



import os
import tempfile

from sqllineage.core.models import SubQuery
from sqllineage.runner import LineageRunner
from sqllineage.utils.constant import LineageLevel
from ..helpers import assert_table_lineage_equal


def test_runner_dummy():
    runner = LineageRunner(
        """insert into tab2 select col1, col2, col3, col4, col5, col6 from tab1;
insert into tab3 select * from tab2""",
        verbose=True,
    )
    assert str(runner)
    assert runner.to_cytoscape() is not None
    assert runner.to_cytoscape(level=LineageLevel.COLUMN) is not None


def test_statements_trim_comment():
    comment = "------------------\n"
    sql = "select * from dual;"
    assert LineageRunner(comment + sql).statements()[0] == sql


def test_silent_mode():
    sql = "begin; select * from dual;"
    LineageRunner(sql, dialect="greenplum", silent_mode=True)._eval()


def test_get_column_lineage_exclude_subquery_inpath():
    v_sql = "insert into ta select b from (select b from tb union all select c from tc ) sub"
    parse = LineageRunner(sql=v_sql)
    for col_tuple in parse.get_column_lineage(exclude_subquery_columns=True):
        for col in col_tuple:
            assert not isinstance(col.parent, SubQuery)


def test_respect_sqlfluff_configuration_file():
    sqlfluff_config = """[sqlfluff:templater:jinja:context]
num_things=456
tbl_name=my_table"""
    cwd = os.getcwd()
    with tempfile.TemporaryDirectory() as tmpdirname:
        try:
            os.chdir(tmpdirname)
            with open(".sqlfluff", "w") as f:
                f.write(sqlfluff_config)
            sql = "SELECT {{ num_things }} FROM {{ tbl_name }} WHERE id > 10 LIMIT 5"
            assert_table_lineage_equal(sql, {"my_table"}, test_sqlparse=False)
        finally:
            os.chdir(cwd)



import pytest

from ...helpers import assert_table_lineage_equal


@pytest.mark.parametrize("dialect", ["exasol", "mysql", "teradata"])
def test_rename_table(dialect: str):
    """
    https://docs.exasol.com/db/latest/sql/rename.htm
    https://dev.mysql.com/doc/refman/8.0/en/rename-table.html
    https://docs.teradata.com/r/Teradata-Database-SQL-Data-Definition-Language-Syntax-and-Examples/December-2015/Table-Statements/RENAME-TABLE
    """
    assert_table_lineage_equal("rename table tab1 to tab2", dialect=dialect)


@pytest.mark.parametrize("dialect", ["mysql"])
def test_rename_tables(dialect: str):
    assert_table_lineage_equal(
        "rename table tab1 to tab2, tab3 to tab4", dialect=dialect
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_refresh_table(dialect: str):
    assert_table_lineage_equal("refresh table tab1", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_cache_table(dialect: str):
    assert_table_lineage_equal("cache table tab1 select * from tab2", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_uncache_table(dialect: str):
    assert_table_lineage_equal("uncache table tab1", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_uncache_table_if_exists(dialect: str):
    assert_table_lineage_equal("uncache table if exists tab1", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_show_create_table(dialect: str):
    assert_table_lineage_equal("show create table tab1", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_add_jar(dialect: str):
    assert_table_lineage_equal("ADD JAR /tmp/SimpleUdf.jar", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_create_function(dialect: str):
    assert_table_lineage_equal(
        """CREATE FUNCTION simple_udf AS 'SimpleUdf'
  USING JAR '/tmp/SimpleUdf.jar'""",
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_create_or_replace_function(dialect: str):
    assert_table_lineage_equal(
        """CREATE OR REPLACE FUNCTION simple_udf AS 'SimpleUdf'
      USING JAR '/tmp/SimpleUdf.jar'""",
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_create_temporary_function(dialect: str):
    assert_table_lineage_equal(
        """CREATE TEMPORARY FUNCTION simple_temp_udf AS 'SimpleUdf'
  USING JAR '/tmp/SimpleUdf.jar'""",
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_create_or_replace_temporary_function(dialect: str):
    assert_table_lineage_equal(
        """CREATE OR REPLACE TEMPORARY FUNCTION simple_temp_udf AS 'SimpleUdf'
  USING JAR '/tmp/SimpleUdf.jar'""",
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_show_functions(dialect: str):
    assert_table_lineage_equal("SHOW FUNCTIONS trim", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_describe_function(dialect: str):
    assert_table_lineage_equal("DESC FUNCTION abs", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_drop_function(dialect: str):
    assert_table_lineage_equal("DROP FUNCTION test_avg", dialect=dialect)


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_set_command(dialect: str):
    assert_table_lineage_equal(
        "SET spark.sql.variable.substitute=false", dialect=dialect
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_set_command_without_property_value(dialect: str):
    assert_table_lineage_equal("SET spark.sql.variable.substitute", dialect=dialect)


@pytest.mark.parametrize("dialect", ["postgres", "redshift"])
def test_analyze_table(dialect: str):
    assert_table_lineage_equal("analyze tab", dialect=dialect)


@pytest.mark.parametrize("dialect", ["postgres", "redshift"])
def test_analyze_table_column(dialect: str):
    assert_table_lineage_equal("analyze tab (col1, col2)", dialect=dialect)



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_select_column_in_subquery():
    sql = """INSERT INTO tab1
SELECT col1
FROM (SELECT col1 FROM tab2) dt"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT col1
FROM (SELECT col1, col2 FROM tab2) dt"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT col1
FROM (SELECT col1 FROM tab2)"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_select_column_in_subquery_with_two_parenthesis():
    sql = """INSERT INTO tab1
SELECT col1
FROM ((SELECT col1 FROM tab2)) dt"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_select_column_in_subquery_with_two_parenthesis_and_blank_in_between():
    sql = """INSERT INTO tab1
SELECT col1
FROM (
(SELECT col1 FROM tab2)
) dt"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_select_column_in_subquery_with_two_parenthesis_and_union():
    sql = """INSERT INTO tab1
SELECT col1
FROM (
    (SELECT col1 FROM tab2)
    UNION ALL
    (SELECT col1 FROM tab3)
) dt"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col1", "tab3"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
        ],
    )


def test_select_column_in_subquery_with_two_parenthesis_and_union_v2():
    sql = """INSERT INTO tab1
SELECT col1
FROM (
    SELECT col1 FROM tab2
    UNION ALL
    SELECT col1 FROM tab3
) dt"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col1", "tab3"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
        ],
    )


def test_window_function_in_subquery():
    sql = """INSERT INTO tab1
SELECT rn FROM (
    SELECT
        row_number() over (partition BY col1, col2) rn
    FROM tab2
) sub
WHERE rn = 1"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("rn", "tab1")),
            (ColumnQualifierTuple("col2", "tab2"), ColumnQualifierTuple("rn", "tab1")),
        ],
    )


def test_select_column_in_subquery_alias_and_qualifier_case_insensitive():
    sql = """INSERT INTO tab1
SELECT DT.col1
FROM (SELECT col1 FROM tab2) dt"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT dt.col1
FROM (SELECT col1 FROM tab2) DT"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT Dt.col1
FROM (SELECT col1 FROM tab2) dT"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )



import pytest

from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_select_column_using_cast():
    sql = """INSERT INTO tab1
SELECT cast(col1 AS timestamp)
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("cast(col1 as timestamp)", "tab1"),
            )
        ],
    )
    sql = """INSERT INTO tab1
SELECT cast(col1 AS timestamp) AS col2
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col2", "tab1"))],
    )


def test_cast_with_comparison():
    sql = """INSERT INTO tab1
SELECT cast(col1 = 1 AS int) col1, col2 = col3 col2
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col2", "tab1"),
            ),
            (
                ColumnQualifierTuple("col3", "tab2"),
                ColumnQualifierTuple("col2", "tab1"),
            ),
        ],
    )


@pytest.mark.parametrize(
    "dtype", ["string", "timestamp", "date", "decimal(18, 0)", "varchar(255)"]
)
def test_cast_to_data_type(dtype: str):
    sql = f"""INSERT INTO tab1
SELECT cast(col1 as {dtype}) AS col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


@pytest.mark.parametrize(
    "dtype", ["string", "timestamp", "date", "decimal(18, 0)", "varchar(255)"]
)
def test_nested_cast_to_data_type(dtype: str):
    sql = f"""INSERT INTO tab1
SELECT cast(cast(col1 AS {dtype}) AS {dtype}) AS col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = f"""INSERT INTO tab1
SELECT cast(cast(cast(cast(cast(col1 AS {dtype}) AS {dtype}) AS {dtype}) AS {dtype}) AS {dtype}) AS col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


@pytest.mark.parametrize(
    "dtype", ["string", "timestamp", "date", "decimal(18, 0)", "varchar(255)"]
)
def test_cast_to_data_type_with_case_when(dtype: str):
    sql = f"""INSERT INTO tab1
SELECT cast(case when col1 > 0 then col2 else col3 end as {dtype}) AS col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col3", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
        ],
    )


def test_cast_using_constant():
    sql = """INSERT INTO tab1
SELECT cast('2012-12-21' AS date) AS col2"""
    assert_column_lineage_equal(sql)


def test_postgres_style_type_cast():
    sql = """INSERT INTO tab1
SELECT col1::timestamp
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


@pytest.mark.parametrize(
    "func",
    [
        "coalesce(col1, 0) as varchar",
        "if(col1 > 100, 100, col1) as varchar",
        "ln(col1) as varchar",
        "conv(col1, 10, 2) as varchar",
        "ln(cast(coalesce(col1, '0') as int)) as varchar",
        "coalesce(col1, 0) as decimal(10, 6)",
    ],
)
def test_column_try_cast_with_func(func: str):
    sql = f"""INSERT INTO tab2
SELECT try_cast({func}) AS col2
FROM tab1"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab1"),
                ColumnQualifierTuple("col2", "tab2"),
            ),
        ],
    )



import pytest

from ...helpers import assert_table_lineage_equal


@pytest.mark.parametrize("dialect", ["mysql"])
def test_update_with_join(dialect: str):
    assert_table_lineage_equal(
        "UPDATE tab1 a INNER JOIN tab2 b ON a.col1=b.col1 SET a.col2=b.col2",
        {"tab2"},
        {"tab1"},
        dialect=dialect,
    )



import itertools
from typing import Dict, List, Optional, Set, Tuple, Union

import networkx as nx
from networkx import DiGraph

from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.core.models import Column, Path, Schema, SubQuery, Table
from sqllineage.utils.constant import EdgeTag, EdgeType, NodeTag

DATASET_CLASSES = (Path, Table)


class ColumnLineageMixin:
    def get_column_lineage(
        self, exclude_path_ending_in_subquery=True, exclude_subquery_columns=False
    ) -> Set[Tuple[Column, ...]]:
        """
        :param exclude_path_ending_in_subquery:  exclude_subquery rename to exclude_path_ending_in_subquery
               exclude column from SubQuery in the ending path
        :param exclude_subquery_columns: exclude column from SubQuery in the path.

        return a list of column tuple :class:`sqllineage.models.Column`
        """
        self.graph: DiGraph  # For mypy attribute checking
        # filter all the column node in the graph
        column_nodes = [n for n in self.graph.nodes if isinstance(n, Column)]
        column_graph = self.graph.subgraph(column_nodes)
        source_columns = {column for column, deg in column_graph.in_degree if deg == 0}
        # if a column lineage path ends at SubQuery, then it should be pruned
        target_columns = {
            node
            for node, deg in column_graph.out_degree
            if isinstance(node, Column) and deg == 0
        }
        if exclude_path_ending_in_subquery:
            target_columns = {
                node for node in target_columns if isinstance(node.parent, Table)
            }
        columns = set()
        for source, target in itertools.product(source_columns, target_columns):
            simple_paths = list(nx.all_simple_paths(self.graph, source, target))
            for path in simple_paths:
                if exclude_subquery_columns:
                    path = [
                        node for node in path if not isinstance(node.parent, SubQuery)
                    ]
                    if len(path) > 1:
                        columns.add(tuple(path))
                else:
                    columns.add(tuple(path))
        return columns


class SubQueryLineageHolder(ColumnLineageMixin):
    """
    SubQuery/Query Level Lineage Result.

    SubQueryLineageHolder will hold attributes like read, write, cte.

    Each of them is a Set[:class:`sqllineage.core.models.Table`].

    This is the most atomic representation of lineage result.
    """

    def __init__(self) -> None:
        self.graph = nx.DiGraph()

    def __or__(self, other):
        self.graph = nx.compose(self.graph, other.graph)
        return self

    def _property_getter(self, prop) -> Set[Union[SubQuery, Table]]:
        return {t for t, attr in self.graph.nodes(data=True) if attr.get(prop) is True}

    def _property_setter(self, value, prop) -> None:
        self.graph.add_node(value, **{prop: True})

    @property
    def read(self) -> Set[Union[SubQuery, Table]]:
        return self._property_getter(NodeTag.READ)

    def add_read(self, value) -> None:
        self._property_setter(value, NodeTag.READ)
        # the same table can be added (in SQL: joined) multiple times with different alias
        if hasattr(value, "alias"):
            self.graph.add_edge(value, value.alias, type=EdgeType.HAS_ALIAS)

    @property
    def write(self) -> Set[Union[SubQuery, Table]]:
        # SubQueryLineageHolder.write can return a single SubQuery or Table, or both when __or__ together.
        # This is different from StatementLineageHolder.write, where Table is the only possibility.
        return self._property_getter(NodeTag.WRITE)

    def add_write(self, value) -> None:
        self._property_setter(value, NodeTag.WRITE)

    @property
    def cte(self) -> Set[SubQuery]:
        return self._property_getter(NodeTag.CTE)  # type: ignore

    def add_cte(self, value) -> None:
        self._property_setter(value, NodeTag.CTE)

    @property
    def write_columns(self) -> List[Column]:
        """
        return a list of columns that write table contains.
        It's either manually added via `add_write_column` if specified in DML
        or automatic added via `add_column_lineage` after parsing from SELECT
        """
        tgt_cols = []
        if tgt_tbl := self._get_target_table():
            tgt_col_with_idx: List[Tuple[Column, int]] = sorted(
                [
                    (col, attr.get(EdgeTag.INDEX, 0))
                    for tbl, col, attr in self.graph.out_edges(tgt_tbl, data=True)
                    if attr["type"] == EdgeType.HAS_COLUMN
                ],
                key=lambda x: x[1],
            )
            tgt_cols = [x[0] for x in tgt_col_with_idx]
        return tgt_cols

    def add_write_column(self, *tgt_cols: Column) -> None:
        """
        in case of DML with column specified, like:

        .. code-block:: sql

            INSERT INTO tab1 (col1, col2)
            SELECT col3, col4

        this method is called to make sure tab1 has column col1 and col2 instead of col3 and col4
        """
        if self.write:
            tgt_tbl = list(self.write)[0]
            for idx, tgt_col in enumerate(tgt_cols):
                tgt_col.parent = tgt_tbl
                if tgt_col in self.write_columns:
                    # for DDL with PARTITIONED BY (col) or CLUSTERED BY (col), column can be added multiple times
                    break
                self.graph.add_edge(
                    tgt_tbl, tgt_col, type=EdgeType.HAS_COLUMN, **{EdgeTag.INDEX: idx}
                )

    def add_column_lineage(self, src: Column, tgt: Column) -> None:
        """
        link source column to target.
        """
        self.graph.add_edge(src, tgt, type=EdgeType.LINEAGE)
        self.graph.add_edge(tgt.parent, tgt, type=EdgeType.HAS_COLUMN)
        if src.parent is not None:
            # starting NetworkX v2.6, None is not allowed as node, see https://github.com/networkx/networkx/pull/4892
            self.graph.add_edge(src.parent, src, type=EdgeType.HAS_COLUMN)

    def get_table_columns(self, table: Union[Table, SubQuery]) -> List[Column]:
        return [
            tgt
            for (src, tgt, edge_type) in self.graph.out_edges(nbunch=table, data="type")
            if edge_type == EdgeType.HAS_COLUMN
            and isinstance(tgt, Column)
            and tgt.raw_name != "*"
        ]

    def expand_wildcard(self, metadata_provider: MetaDataProvider) -> None:
        if tgt_table := self._get_target_table():
            for column in self.write_columns:
                if column.raw_name == "*":
                    tgt_wildcard = column
                    for src_wildcard in self.get_source_columns(tgt_wildcard):
                        if source_table := src_wildcard.parent:
                            src_table_columns = []
                            if isinstance(source_table, SubQuery):
                                # the columns of SubQuery can be inferred from graph
                                src_table_columns = self.get_table_columns(source_table)
                            elif isinstance(source_table, Table) and metadata_provider:
                                # search by metadata service
                                src_table_columns = metadata_provider.get_table_columns(
                                    source_table
                                )
                            if src_table_columns:
                                self._replace_wildcard(
                                    tgt_table,
                                    src_table_columns,
                                    tgt_wildcard,
                                    src_wildcard,
                                )

    def get_alias_mapping_from_table_group(
        self, table_group: List[Union[Path, Table, SubQuery]]
    ) -> Dict[str, Union[Path, Table, SubQuery]]:
        """
        A table can be referred to as alias, table name, or database_name.table_name, create the mapping here.
        For SubQuery, it's only alias then.
        """
        return {
            **{
                tgt: src
                for src, tgt, attr in self.graph.edges(data=True)
                if attr.get("type") == EdgeType.HAS_ALIAS and src in table_group
            },
            **{
                table.raw_name: table
                for table in table_group
                if isinstance(table, Table)
            },
            **{str(table): table for table in table_group if isinstance(table, Table)},
        }

    def _get_target_table(self) -> Optional[Union[SubQuery, Table]]:
        table = None
        if write_only := self.write.difference(self.read):
            table = next(iter(write_only))
        return table

    def get_source_columns(self, node: Column) -> List[Column]:
        return [
            src
            for (src, tgt, edge_type) in self.graph.in_edges(nbunch=node, data="type")
            if edge_type == EdgeType.LINEAGE and isinstance(src, Column)
        ]

    def _replace_wildcard(
        self,
        tgt_table: Union[Table, SubQuery],
        src_table_columns: List[Column],
        tgt_wildcard: Column,
        src_wildcard: Column,
    ) -> None:
        target_columns = self.get_table_columns(tgt_table)
        for src_col in src_table_columns:
            new_column = Column(src_col.raw_name)
            new_column.parent = tgt_table
            if new_column in target_columns or src_col.raw_name == "*":
                continue
            self.graph.add_edge(tgt_table, new_column, type=EdgeType.HAS_COLUMN)
            self.graph.add_edge(src_col.parent, src_col, type=EdgeType.HAS_COLUMN)
            self.graph.add_edge(src_col, new_column, type=EdgeType.LINEAGE)
        # remove wildcard
        if self.graph.has_node(tgt_wildcard):
            self.graph.remove_node(tgt_wildcard)
        if self.graph.has_node(src_wildcard):
            self.graph.remove_node(src_wildcard)


class StatementLineageHolder(SubQueryLineageHolder, ColumnLineageMixin):
    """
    Statement Level Lineage Result.

    Based on SubQueryLineageHolder, StatementLineageHolder holds extra attributes like drop and rename

    For drop, it is a Set[:class:`sqllineage.core.models.Table`].

    For rename, it a Set[Tuple[:class:`sqllineage.core.models.Table`, :class:`sqllineage.core.models.Table`]],
    with the first table being original table before renaming and the latter after renaming.
    """

    def __str__(self):
        return "\n".join(
            f"table {attr}: {sorted(getattr(self, attr), key=lambda x: str(x)) if getattr(self, attr) else '[]'}"
            for attr in ["read", "write", "cte", "drop", "rename"]
        )

    def __repr__(self):
        return str(self)

    @property
    def read(self) -> Set[Table]:  # type: ignore
        return {t for t in super().read if isinstance(t, DATASET_CLASSES)}

    @property
    def write(self) -> Set[Table]:  # type: ignore
        return {t for t in super().write if isinstance(t, DATASET_CLASSES)}

    @property
    def drop(self) -> Set[Table]:
        return self._property_getter(NodeTag.DROP)  # type: ignore

    def add_drop(self, value) -> None:
        self._property_setter(value, NodeTag.DROP)

    @property
    def rename(self) -> Set[Tuple[Table, Table]]:
        return {
            (src, tgt)
            for src, tgt, attr in self.graph.edges(data=True)
            if attr.get("type") == EdgeType.RENAME
        }

    def add_rename(self, src: Table, tgt: Table) -> None:
        self.graph.add_edge(src, tgt, type=EdgeType.RENAME)

    @staticmethod
    def of(holder: SubQueryLineageHolder) -> "StatementLineageHolder":
        stmt_holder = StatementLineageHolder()
        stmt_holder.graph = holder.graph
        return stmt_holder


class SQLLineageHolder(ColumnLineageMixin):
    def __init__(self, graph: DiGraph):
        """
        The combined lineage result in representation of Directed Acyclic Graph.

        :param graph: the Directed Acyclic Graph holding all the combined lineage result.
        """
        self.graph = graph
        self._selfloop_tables = self.__retrieve_tag_tables(NodeTag.SELFLOOP)
        self._sourceonly_tables = self.__retrieve_tag_tables(NodeTag.SOURCE_ONLY)
        self._targetonly_tables = self.__retrieve_tag_tables(NodeTag.TARGET_ONLY)

    @property
    def table_lineage_graph(self) -> DiGraph:
        """
        The table level DiGraph held by SQLLineageHolder
        """
        table_nodes = [n for n in self.graph.nodes if isinstance(n, DATASET_CLASSES)]
        return self.graph.subgraph(table_nodes)

    @property
    def column_lineage_graph(self) -> DiGraph:
        """
        The column level DiGraph held by SQLLineageHolder
        """
        column_nodes = [n for n in self.graph.nodes if isinstance(n, Column)]
        return self.graph.subgraph(column_nodes)

    @property
    def source_tables(self) -> Set[Table]:
        """
        a list of source :class:`sqllineage.core.models.Table`
        """
        source_tables = {
            table for table, deg in self.table_lineage_graph.in_degree if deg == 0
        }.intersection(
            {table for table, deg in self.table_lineage_graph.out_degree if deg > 0}
        )
        source_tables |= self._selfloop_tables
        source_tables |= self._sourceonly_tables
        return source_tables

    @property
    def target_tables(self) -> Set[Table]:
        """
        a list of target :class:`sqllineage.core.models.Table`
        """
        target_tables = {
            table for table, deg in self.table_lineage_graph.out_degree if deg == 0
        }.intersection(
            {table for table, deg in self.table_lineage_graph.in_degree if deg > 0}
        )
        target_tables |= self._selfloop_tables
        target_tables |= self._targetonly_tables
        return target_tables

    @property
    def intermediate_tables(self) -> Set[Table]:
        """
        a list of intermediate :class:`sqllineage.core.models.Table`
        """
        intermediate_tables = {
            table for table, deg in self.table_lineage_graph.in_degree if deg > 0
        }.intersection(
            {table for table, deg in self.table_lineage_graph.out_degree if deg > 0}
        )
        intermediate_tables -= self.__retrieve_tag_tables(NodeTag.SELFLOOP)
        return intermediate_tables

    def __retrieve_tag_tables(self, tag) -> Set[Union[Path, Table]]:
        return {
            table
            for table, attr in self.graph.nodes(data=True)
            if attr.get(tag) is True and isinstance(table, DATASET_CLASSES)
        }

    @staticmethod
    def _build_digraph(
        metadata_provider: MetaDataProvider, *args: StatementLineageHolder
    ) -> DiGraph:
        g = DiGraph()
        for holder in args:
            g = nx.compose(g, holder.graph)
            if holder.drop:
                for table in holder.drop:
                    if g.has_node(table) and g.degree[table] == 0:
                        g.remove_node(table)
            elif holder.rename:
                for table_old, table_new in holder.rename:
                    g = nx.relabel_nodes(g, {table_old: table_new})
                    g.remove_edge(table_new, table_new)
                    if g.degree[table_new] == 0:
                        g.remove_node(table_new)
            else:
                read, write = holder.read, holder.write
                if len(read) > 0 and len(write) == 0:
                    # source only table comes from SELECT statement
                    nx.set_node_attributes(
                        g, {table: True for table in read}, NodeTag.SOURCE_ONLY
                    )
                elif len(read) == 0 and len(write) > 0:
                    # target only table comes from case like: 1) INSERT/UPDATE constant values; 2) CREATE TABLE
                    nx.set_node_attributes(
                        g, {table: True for table in write}, NodeTag.TARGET_ONLY
                    )
                else:
                    for source, target in itertools.product(read, write):
                        g.add_edge(source, target, type=EdgeType.LINEAGE)
        nx.set_node_attributes(
            g,
            {table: True for table in {e[0] for e in nx.selfloop_edges(g)}},
            NodeTag.SELFLOOP,
        )
        # find all the columns that we can't assign accurately to a parent table (with multiple parent candidates)
        unresolved_cols = [
            (s, t)
            for s, t in g.edges
            if isinstance(s, Column) and len(s.parent_candidates) > 1
        ]
        for unresolved_col, tgt_col in unresolved_cols:
            # check if there's only one parent candidate contains the column with same name
            src_cols = []
            # check if source column exists in graph (either from subquery or from table created in prev statement)
            for parent in unresolved_col.parent_candidates:
                src_col = Column(unresolved_col.raw_name)
                src_col.parent = parent
                if g.has_edge(parent, src_col):
                    src_cols.append(src_col)
            # if not in graph, check if defined in table schema by metadata service
            if len(src_cols) == 0 and bool(metadata_provider):
                for parent in unresolved_col.parent_candidates:
                    if (
                        isinstance(parent, Table)
                        and str(parent.schema) != Schema.unknown
                    ):
                        columns = metadata_provider.get_table_columns(parent)
                        for src_col in columns:
                            if unresolved_col.raw_name == src_col.raw_name:
                                src_cols.append(src_col)

            # Multiple sources is a correct case for JOIN with USING
            # It incorrect for JOIN with ON, but sql without specifying an alias in this case will be invalid
            for src_col in src_cols:
                g.add_edge(src_col, tgt_col, type=EdgeType.LINEAGE)
            if len(src_cols) > 0:
                # only delete unresolved column when it's resolved
                g.remove_edge(unresolved_col, tgt_col)

        # when unresolved column got resolved, it will be orphan node, and we can remove it
        for node in [n for n, deg in g.degree if deg == 0]:
            if isinstance(node, Column) and len(node.parent_candidates) > 1:
                g.remove_node(node)
        return g

    @staticmethod
    def of(metadata_provider, *args: StatementLineageHolder) -> "SQLLineageHolder":
        """
        To assemble multiple :class:`sqllineage.core.holders.StatementLineageHolder` into
        :class:`sqllineage.core.holders.SQLLineageHolder`
        """
        g = SQLLineageHolder._build_digraph(metadata_provider, *args)
        return SQLLineageHolder(g)



from ....helpers import assert_table_lineage_equal


def test_drop_tmp_tab_after_create():
    sql = """CREATE TABLE tab_a AS
SELECT *
FROM tab_b;
INSERT INTO tab_c
SELECT *
FROM tab_a;
DROP TABLE tab_a;"""
    assert_table_lineage_equal(sql, {"tab_b"}, {"tab_c"})


def test_new_create_tab_as_tmp_table():
    sql = """CREATE TABLE tab_a AS
SELECT *
FROM tab_b;
CREATE TABLE tab_c AS
SELECT *
FROM tab_a;"""
    assert_table_lineage_equal(sql, {"tab_b"}, {"tab_c"})


def test_create_after_drop():
    assert_table_lineage_equal(
        "DROP TABLE IF EXISTS tab1; CREATE TABLE IF NOT EXISTS tab1 AS SELECT 1",
        None,
        {"tab1"},
    )


def test_drop_after_create():
    assert_table_lineage_equal(
        "CREATE TABLE IF NOT EXISTS tab1 AS SELECT 1; DROP TABLE IF EXISTS tab1",
        None,
        None,
    )


def test_alter_target_table_name():
    assert_table_lineage_equal(
        "INSERT INTO tab1 SELECT * FROM tab2; ALTER TABLE tab1 RENAME TO tab3;",
        {"tab2"},
        {"tab3"},
    )



import os
from unittest.mock import patch

import pytest

from sqllineage import SQLPARSE_DIALECT
from sqllineage.exceptions import (
    InvalidSyntaxException,
    SQLLineageException,
    UnsupportedStatementException,
)
from sqllineage.runner import LineageRunner


def test_select_without_table():
    with pytest.raises(SQLLineageException):
        LineageRunner("select * from where foo='bar'")._eval()
    with pytest.raises(SQLLineageException):
        LineageRunner("select * from where foo='bar'", dialect=SQLPARSE_DIALECT)._eval()


def test_full_unparsable_query_in_sqlfluff():
    with pytest.raises(InvalidSyntaxException):
        LineageRunner("WRONG SELECT FROM tab1")._eval()


def test_partial_unparsable_query_in_sqlfluff():
    with pytest.raises(InvalidSyntaxException):
        LineageRunner("SELECT * FROM tab1 AS FULL FULL OUTER JOIN tab2")._eval()


def test_partial_unparsable_query_in_sqlfluff_with_tsql_batch():
    sql = """SELECT *
INTO tgt
FROM tab1 src1 AS src1
CROSS JOIN tab2 AS src2"""
    with pytest.raises(InvalidSyntaxException):
        LineageRunner(sql, dialect="tsql")._eval()


def test_unsupported_query_type_in_sqlfluff():
    with pytest.raises(UnsupportedStatementException):
        LineageRunner("CREATE UNIQUE INDEX title_idx ON films (title)")._eval()


def test_deprecation_warning_in_sqlparse():
    with pytest.warns(DeprecationWarning):
        LineageRunner("SELECT * FROM DUAL", dialect=SQLPARSE_DIALECT)._eval()


def test_syntax_warning_no_semicolon_in_tsql():
    with pytest.warns(SyntaxWarning):
        LineageRunner(
            """SELECT * FROM foo
SELECT * FROM bar""",
            dialect="tsql",
        )._eval()


@patch.dict(os.environ, {"SQLLINEAGE_TSQL_NO_SEMICOLON": "TRUE"})
def test_user_warning_enable_tsql_no_semicolon_with_other_dialect():
    with pytest.warns(UserWarning):
        LineageRunner(
            """SELECT * FROM foo;
SELECT * FROM bar""",
        )._eval()



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_column_update_with_single_table():
    sql = """
    update
        public.tgt_tbl1 as t
    set
        name  = s.name,
        email = s.address
    from
        public.src_tbl1 as s
    where
        s.id = t.id
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("address", "public.src_tbl1"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )


def test_column_update_with_cte():
    sql = """
    with
        s as (
            select name, address from public.src_tbl1
        )
    update
        public.tgt_tbl1 as t
    set
        name  = s.name,
        email = s.address
    from
        s
    where
        s.id = t.id
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("address", "public.src_tbl1"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )


def test_column_update_with_multi_tables():
    sql = """
    update
        public.tgt_tbl1 as t
    set
        name  = s1.name,
        email = s2.email
    from
        public.src_tbl1 as s1
        join
            public.src_tbl2 as s2
        on
            s1.id = s2.id
    where
        s1.id = t.id
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl2"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )
    sql = """
    update
        public.tgt_tbl1 as t
    set
        name = s1.name,
        email = s2.email
    from
        public.src_tbl1 as s1,
        public.src_tbl2 as s2
    where
        s1.id = s2.id and
        s1.id = t.id
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl2"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )


def test_column_update_with_subquery():
    sql = """
    update
        public.tgt_tbl1 as t
    set
        name  = s.name,
        email = s.email
    from
        (
            select
                s1.id,
                s1.name,
                s2.email
            from
                public.src_tbl1 as s1
                join
                    public.src_tbl2 as s2
                on
                    s1.id = s2.id
        ) as s
    where
        s.id = t.id
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl2"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )



from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import StatementLineageHolder
from sqllineage.core.models import Path
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.core.parser.sqlfluff.utils import (
    find_from_expression_element,
    list_child_segments,
)
from sqllineage.utils.entities import AnalyzerContext
from sqllineage.utils.helpers import escape_identifier_name


class CopyExtractor(BaseExtractor):
    """
    Copy statement lineage extractor
    """

    SUPPORTED_STMT_TYPES = [
        "copy_statement",
        "copy_into_table_statement",
    ]

    def extract(
        self, statement: BaseSegment, context: AnalyzerContext
    ) -> StatementLineageHolder:
        holder = StatementLineageHolder()
        src_flag = tgt_flag = False
        for segment in list_child_segments(statement):
            if segment.type == "from_clause":
                if from_expression_element := find_from_expression_element(segment):
                    for table_expression in from_expression_element.get_children(
                        "table_expression"
                    ):
                        if storage_location := table_expression.get_child(
                            "storage_location"
                        ):
                            holder.add_read(Path(storage_location.raw))
            elif segment.type == "keyword":
                if segment.raw_upper in ["COPY", "INTO"]:
                    tgt_flag = True
                elif segment.raw_upper == "FROM":
                    src_flag = True
                continue

            if tgt_flag:
                if table := self.find_table(segment):
                    holder.add_write(table)
                tgt_flag = False
            if src_flag:
                if segment.type in ["literal", "storage_location"]:
                    path = Path(escape_identifier_name(segment.raw))
                    holder.add_read(path)
                src_flag = False

        return holder



from unittest.mock import Mock

from sqlfluff.core import Linter

from sqllineage.core.parser.sqlfluff.models import SqlFluffColumn
from sqllineage.core.parser.sqlfluff.utils import find_from_expression_element


def test_column_extract_source_columns():
    segment_mock = Mock()
    segment_mock.type = ""
    assert [] == SqlFluffColumn._extract_source_columns(segment_mock)


def test_return_none_find_from_expression_element():
    file_segment = Linter(dialect="ansi").parse_string("TRUNCATE TABLE tab").tree
    assert find_from_expression_element(file_segment) is None



from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.core.parser.sqlfluff.extractors.create_insert import (
    CreateInsertExtractor,
)
from sqllineage.core.parser.sqlfluff.extractors.select import SelectExtractor
from sqllineage.core.parser.sqlfluff.extractors.update import UpdateExtractor
from sqllineage.core.parser.sqlfluff.models import SqlFluffSubQuery
from sqllineage.core.parser.sqlfluff.utils import list_child_segments
from sqllineage.utils.entities import AnalyzerContext


class CteExtractor(BaseExtractor):
    """
    CTE queries lineage extractor
    """

    SUPPORTED_STMT_TYPES = ["with_compound_statement"]

    def extract(
        self,
        statement: BaseSegment,
        context: AnalyzerContext,
    ) -> SubQueryLineageHolder:
        holder = self._init_holder(context)
        subqueries = []
        for segment in list_child_segments(statement):
            if segment.type in ["select_statement", "set_expression"]:
                holder |= self.delegate_to(
                    SelectExtractor,
                    segment,
                    AnalyzerContext(
                        cte=holder.cte,
                        write=holder.write,
                        write_columns=holder.write_columns,
                    ),
                )
            elif segment.type == "insert_statement":
                holder |= self.delegate_to(
                    CreateInsertExtractor, segment, AnalyzerContext(cte=holder.cte)
                )
            elif segment.type == "update_statement":
                holder |= self.delegate_to(
                    UpdateExtractor, segment, AnalyzerContext(cte=holder.cte)
                )
            elif segment.type == "common_table_expression":
                alias = None
                sub_segments = list_child_segments(segment)
                for sub_segment in sub_segments:
                    if sub_segment.type == "identifier":
                        alias = sub_segment.raw
                    elif sub_segment.type == "bracketed":
                        for sq in self.list_subquery(sub_segment):
                            sq.alias = alias
                            subqueries.append(sq)
                        holder.add_cte(SqlFluffSubQuery.of(sub_segment, alias))

        self.extract_subquery(subqueries, holder)

        return holder



from sqlparse.sql import Comparison, Function, Identifier, Token
from sqlparse.tokens import Literal, Number

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.models import Path
from sqllineage.core.parser.sqlparse.handlers.base import NextTokenBaseHandler
from sqllineage.core.parser.sqlparse.models import SqlParseTable
from sqllineage.exceptions import SQLLineageException


class TargetHandler(NextTokenBaseHandler):
    """Target Table Handler."""

    TARGET_TABLE_TOKENS = (
        "INTO",
        "OVERWRITE",
        "TABLE",
        "VIEW",
        "UPDATE",
        "COPY",
        "DIRECTORY",
    )

    def _indicate(self, token: Token) -> bool:
        if (
            self.indicator is True
            and token.is_keyword
            and token.normalized in ("IF", "NOT", "EXISTS")
        ):
            # special handling for CREATE TABLE IF NOT EXISTS
            return True
        else:
            return token.normalized in self.TARGET_TABLE_TOKENS

    def _handle(self, token: Token, holder: SubQueryLineageHolder) -> None:
        if isinstance(token, Function):
            # insert into tab (col1, col2) values (val1, val2); Here tab (col1, col2) will be parsed as Function
            # referring https://github.com/andialbrecht/sqlparse/issues/483 for further information
            if not isinstance(token.token_first(skip_cm=True), Identifier):
                raise SQLLineageException(
                    "An Identifier is expected, got %s[value: %s] instead."
                    % (type(token).__name__, token)
                )
            holder.add_write(SqlParseTable.of(token.token_first(skip_cm=True)))
        elif isinstance(token, Comparison):
            # create table tab1 like tab2, tab1 like tab2 will be parsed as Comparison
            # referring https://github.com/andialbrecht/sqlparse/issues/543 for further information
            if not (
                isinstance(token.left, Identifier)
                and isinstance(token.right, Identifier)
            ):
                raise SQLLineageException(
                    "An Identifier is expected, got %s[value: %s] instead."
                    % (type(token).__name__, token)
                )
            holder.add_write(SqlParseTable.of(token.left))
            holder.add_read(SqlParseTable.of(token.right))
        elif token.ttype == Literal.String.Single:
            holder.add_write(Path(token.value))
        elif isinstance(token, Identifier):
            if token.token_first(skip_cm=True).ttype is Number.Integer:
                # Special Handling for Spark Bucket Table DDL
                pass
            else:
                holder.add_write(SqlParseTable.of(token))



import importlib
import os
import pkgutil

# import each module so that BaseExtractor's __subclasses__ will work
for module in pkgutil.iter_modules([os.path.dirname(__file__)]):
    importlib.import_module(__name__ + "." + module.name)



from ...helpers import assert_table_lineage_equal


def test_update():
    assert_table_lineage_equal(
        "UPDATE tab1 SET col1='val1' WHERE col2='val2'", None, {"tab1"}
    )


def test_update_from():
    assert_table_lineage_equal(
        """UPDATE tab2
SET tab2.col2 = tab1.col2 FROM tab1
WHERE tab2.col1 = tab1.col1""",
        {"tab1"},
        {"tab2"},
    )



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_select_column_using_expression():
    sql = """INSERT INTO tab1
SELECT col1 + col2
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1 + col2", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col1 + col2", "tab1"),
            ),
        ],
    )
    sql = """INSERT INTO tab1
SELECT col1 + col2 AS col3
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
    )


def test_select_column_multiply_expression():
    sql = """INSERT INTO tab1
SELECT col1 * 1 AS col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
        ],
    )


def test_select_column_using_expression_in_parenthesis():
    sql = """INSERT INTO tab1
SELECT (col1 + col2)
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("(col1 + col2)", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("(col1 + col2)", "tab1"),
            ),
        ],
    )
    sql = """INSERT INTO tab1
SELECT (col1 + col2) AS col3
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
    )


def test_select_column_using_boolean_expression_in_parenthesis():
    sql = """INSERT INTO tab1
SELECT (col1 > 0 AND col2 > 0) AS col3
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
    )


def test_select_column_using_expression_with_table_qualifier_without_column_alias():
    sql = """INSERT INTO tab1
SELECT a.col1 + a.col2 + a.col3 + a.col4
FROM tab2 a"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("a.col1 + a.col2 + a.col3 + a.col4", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("a.col1 + a.col2 + a.col3 + a.col4", "tab1"),
            ),
            (
                ColumnQualifierTuple("col3", "tab2"),
                ColumnQualifierTuple("a.col1 + a.col2 + a.col3 + a.col4", "tab1"),
            ),
            (
                ColumnQualifierTuple("col4", "tab2"),
                ColumnQualifierTuple("a.col1 + a.col2 + a.col3 + a.col4", "tab1"),
            ),
        ],
    )


def test_subquery_expression_without_source_table():
    assert_column_lineage_equal(
        """INSERT INTO foo
SELECT (SELECT col1 + col2 AS result) AS sum_result
FROM bar""",
        [
            (
                ColumnQualifierTuple("col1", "bar"),
                ColumnQualifierTuple("sum_result", "foo"),
            ),
            (
                ColumnQualifierTuple("col2", "bar"),
                ColumnQualifierTuple("sum_result", "foo"),
            ),
        ],
    )



import os
from pathlib import Path
from unittest.mock import patch

import pytest

from sqllineage.cli import main
from sqllineage.config import SQLLineageConfig


@patch("socketserver.BaseServer.serve_forever")
def test_cli_dummy(_):
    main([])
    main(["-e", "select * from dual"])
    main(["-e", "insert into foo select * from dual", "-l", "column"])
    for dirname, _, files in os.walk(SQLLineageConfig.DIRECTORY):
        if len(files) > 0:
            sql_file = str(Path(dirname).joinpath(Path(files[0])))
            main(["-f", sql_file])
            main(["-e", "select * from dual", "-f", sql_file])
            main(["-f", sql_file, "-g"])
            main(["-f", sql_file, "--silent_mode"])
            main(["-f", sql_file, "--sqlalchemy_url=sqlite:///:memory:"])
            break
    main(["-g"])
    main(["-ds"])
    main(
        [
            "-e",
            "insert overwrite table tab1 select * from tab1 union select * from tab2",
            "-g",
        ]
    )


def test_file_exception():
    for args in (["-f", str(Path().absolute())], ["-f", "nonexist_file"]):
        with pytest.raises(SystemExit) as e:
            main(args)
        assert e.value.code == 1


@patch("builtins.open", side_effect=PermissionError())
def test_file_permission_error(_):
    with pytest.raises(SystemExit) as e:
        main(["-f", __file__])
    assert e.value.code == 1



from ...helpers import assert_table_lineage_equal


def test_select():
    assert_table_lineage_equal("SELECT col1 FROM tab1", {"tab1"})


def test_select_with_schema():
    assert_table_lineage_equal("SELECT col1 FROM schema1.tab1", {"schema1.tab1"})


def test_select_with_schema_and_database():
    assert_table_lineage_equal(
        "SELECT col1 FROM db1.schema1.tbl1", {"db1.schema1.tbl1"}
    )


def test_select_multi_line():
    assert_table_lineage_equal(
        """SELECT col1 FROM
tab1""",
        {"tab1"},
    )


def test_select_asterisk():
    assert_table_lineage_equal("SELECT * FROM tab1", {"tab1"})


def test_select_value():
    assert_table_lineage_equal("SELECT 1")


def test_select_function():
    assert_table_lineage_equal("SELECT NOW()")


def test_select_trim_function_with_from_keyword():
    assert_table_lineage_equal("SELECT trim(BOTH '  ' FROM '  abc  ')")


def test_select_trim_function_with_from_keyword_from_source_table():
    assert_table_lineage_equal("SELECT trim(BOTH '  ' FROM col1) FROM tab1", {"tab1"})


def test_select_with_where():
    assert_table_lineage_equal(
        "SELECT * FROM tab1 WHERE col1 > val1 AND col2 = 'val2'", {"tab1"}
    )


def test_select_with_parenthesized_where():
    assert_table_lineage_equal(
        "SELECT * FROM tab1 WHERE (col1 > val1 AND col2 = 'val2')", {"tab1"}
    )


def test_select_with_comment():
    assert_table_lineage_equal("SELECT -- comment1\n col1 FROM tab1", {"tab1"})


def test_select_with_comment_after_from():
    assert_table_lineage_equal("SELECT col1\nFROM  -- comment\ntab1", {"tab1"})


def test_select_with_comment_after_join():
    assert_table_lineage_equal(
        "SELECT * FROM tab1 JOIN --comment\ntab2 ON tab1.x = tab2.x", {"tab1", "tab2"}
    )


def test_select_keyword_as_column_alias():
    # here "as" is the column alias
    assert_table_lineage_equal('SELECT 1 "as" FROM tab1', {"tab1"})
    # the following is hive specific, MySQL doesn't allow this syntax. As of now, we don't test against it
    # assert_table_lineage_equal("SELECT 1 as FROM tab1", {"tab1"})


def test_select_with_table_alias():
    assert_table_lineage_equal("SELECT 1 FROM tab1 AS alias1", {"tab1"})


def test_select_count():
    assert_table_lineage_equal("SELECT count(*) FROM tab1", {"tab1"})


def test_select_parenthesized_from_table():
    sql = "SELECT * FROM (tab1)"
    assert_table_lineage_equal(sql, {"tab1"})


def test_select_parenthesized_from_table_join():
    sql = """SELECT tab1.id
FROM (tab1
    LEFT JOIN tab2 ON tab1.id = tab2.id)"""
    assert_table_lineage_equal(sql, {"tab1", "tab2"})


def test_select_parenthesized_from_table_join_using():
    sql = """SELECT tab1.id
FROM (tab1
    LEFT JOIN tab2 USING (id))"""
    assert_table_lineage_equal(sql, {"tab1", "tab2"})


def test_select_parenthesized_from_table_join_recursive():
    sql = """SELECT tab1.id
FROM ((tab1
    LEFT JOIN tab2 ON tab1.id = tab2.id)
    LEFT JOIN tab3 ON tab1.id = tab3.id)
    """
    assert_table_lineage_equal(sql, {"tab1", "tab2", "tab3"})
    sql = """SELECT tab1.id
FROM (((tab1
    LEFT JOIN tab2 ON tab1.id = tab2.id)
    LEFT JOIN tab3 ON tab1.id = tab3.id)
    LEFT JOIN tab4 ON tab1.id = tab4.id)
    """
    assert_table_lineage_equal(sql, {"tab1", "tab2", "tab3", "tab4"})
    sql = """SELECT tab1.id
FROM ((((tab1
    LEFT JOIN tab2 ON tab1.id = tab2.id)
    LEFT JOIN tab3 ON tab1.id = tab3.id)
    LEFT JOIN tab4 ON tab1.id = tab4.id)
    LEFT JOIN tab5 ON tab1.id = tab5.id)
    """
    assert_table_lineage_equal(sql, {"tab1", "tab2", "tab3", "tab4", "tab5"})
    sql = """SELECT tab1.id
FROM (((((tab1
    LEFT JOIN tab2 ON tab1.id = tab2.id)
    LEFT JOIN tab3 ON tab1.id = tab3.id)
    LEFT JOIN tab4 ON tab1.id = tab4.id)
    LEFT JOIN tab5 ON tab1.id = tab5.id)
    LEFT JOIN tab6 ON tab1.id = tab6.id)
        """
    assert_table_lineage_equal(sql, {"tab1", "tab2", "tab3", "tab4", "tab5", "tab6"})


def test_select_subquery():
    assert_table_lineage_equal("SELECT col1 FROM (SELECT col1 FROM tab1) dt", {"tab1"})
    # with an extra space
    assert_table_lineage_equal("SELECT col1 FROM ( SELECT col1 FROM tab1) dt", {"tab1"})


def test_select_subquery_with_two_parenthesis():
    assert_table_lineage_equal(
        "SELECT col1 FROM ((SELECT col1 FROM tab1)) dt", {"tab1"}
    )


def test_select_subquery_with_more_parenthesis():
    assert_table_lineage_equal(
        "SELECT col1 FROM (((((((SELECT col1 FROM tab1))))))) dt", {"tab1"}
    )


def test_select_subquery_in_case():
    assert_table_lineage_equal(
        """SELECT
CASE WHEN (SELECT count(*) FROM tab1 WHERE col1 = 'tab2') = 1 THEN (SELECT count(*) FROM tab2) ELSE 0 END AS cnt""",
        {"tab1", "tab2"},
    )
    assert_table_lineage_equal(
        """SELECT
CASE WHEN 1 = (SELECT count(*) FROM tab1 WHERE col1 = 'tab2') THEN (SELECT count(*) FROM tab2) ELSE 0 END AS cnt""",
        {"tab1", "tab2"},
    )
    assert_table_lineage_equal(
        """
        select
            1,
            case
                when 1 = ( select count(*) from tab1 where col1 = 'tab2' )
                    then ( select count(*) from tab2 )
                else 0
            end as cnt
        """,
        {"tab1", "tab2"},
    )


def test_select_subquery_without_alias():
    """this syntax is valid in SparkSQL, not for MySQL"""
    assert_table_lineage_equal("SELECT col1 FROM (SELECT col1 FROM tab1)", {"tab1"})


def test_select_subquery_in_where_clause():
    assert_table_lineage_equal(
        """SELECT col1
FROM tab1
WHERE col1 IN (SELECT max(col1) FROM tab2)""",
        {"tab1", "tab2"},
    )


def test_select_inner_join():
    assert_table_lineage_equal("SELECT * FROM tab1 INNER JOIN tab2", {"tab1", "tab2"})


def test_select_join():
    assert_table_lineage_equal("SELECT * FROM tab1 JOIN tab2", {"tab1", "tab2"})


def test_select_left_join():
    assert_table_lineage_equal("SELECT * FROM tab1 LEFT JOIN tab2", {"tab1", "tab2"})


def test_select_left_join_with_extra_space_in_middle():
    assert_table_lineage_equal("SELECT * FROM tab1 LEFT  JOIN tab2", {"tab1", "tab2"})


def test_select_right_join():
    assert_table_lineage_equal("SELECT * FROM tab1 RIGHT JOIN tab2", {"tab1", "tab2"})


def test_select_full_outer_join():
    assert_table_lineage_equal(
        "SELECT * FROM tab1 FULL OUTER JOIN tab2", {"tab1", "tab2"}
    )


def test_select_cross_join():
    assert_table_lineage_equal("SELECT * FROM tab1 CROSS JOIN tab2", {"tab1", "tab2"})


def test_select_cross_join_with_on():
    assert_table_lineage_equal(
        "SELECT * FROM tab1 CROSS JOIN tab2 ON tab1.col1 = tab2.col2", {"tab1", "tab2"}
    )


def test_select_join_with_subquery():
    assert_table_lineage_equal(
        "SELECT col1 FROM tab1 AS a LEFT JOIN tab2 AS b ON a.id=b.tab1_id "
        "WHERE col1 = (SELECT col1 FROM tab2 WHERE id = 1)",
        {"tab1", "tab2"},
    )


def test_select_join_in_ansi89_syntax():
    assert_table_lineage_equal("SELECT * FROM tab1 a, tab2 b", {"tab1", "tab2"})


def test_select_join_in_ansi89_syntax_with_subquery():
    assert_table_lineage_equal(
        "SELECT * FROM (SELECT * FROM tab1) a, (SELECT * FROM tab2) b", {"tab1", "tab2"}
    )


def test_select_group_by():
    assert_table_lineage_equal(
        "SELECT col1, col2 FROM tab1 GROUP BY col1, col2", {"tab1"}
    )


def test_select_group_by_ordinal():
    assert_table_lineage_equal("SELECT col1, col2 FROM tab1 GROUP BY 1, 2", {"tab1"})


def test_select_from_values():
    assert_table_lineage_equal("SELECT * FROM (VALUES (1, 2))")


def test_select_from_values_newline():
    assert_table_lineage_equal("SELECT * FROM (\nVALUES (1, 2))")


def test_select_from_values_with_alias():
    assert_table_lineage_equal("SELECT * FROM (VALUES (1, 2)) AS t(col1, col2)")


def test_select_from_unnest():
    # unnest function is Presto specific
    assert_table_lineage_equal(
        "SELECT student, score FROM tests CROSS JOIN UNNEST(scores) AS t (score)",
        {"tests"},
    )


def test_select_from_unnest_parsed_as_keyword():
    # an extra space after UNNEST changes the AST structure
    assert_table_lineage_equal(
        "SELECT student, score FROM tests CROSS JOIN UNNEST (scores) AS t (score)",
        {"tests"},
    )


def test_table_name_case_insensitive():
    assert_table_lineage_equal(
        """SELECT * FROM tab_b
UNION ALL
SELECT * FROM TAB_B""",
        {"tab_b"},
    )


def test_select_union_all():
    sql = """SELECT col1
FROM tab1
UNION ALL
SELECT col1
FROM tab2
UNION ALL
SELECT col1
FROM tab3
ORDER BY col1"""
    assert_table_lineage_equal(
        sql,
        {"tab1", "tab2", "tab3"},
    )


def test_select_union_all_with_nested_subquery():
    sql = """SELECT sq1.id
FROM (SELECT id
      FROM tab1) sq1
UNION ALL
SELECT sq2.id
FROM (SELECT id
      FROM tab2) sq2"""
    assert_table_lineage_equal(sql, {"tab1", "tab2"})


def test_non_reserved_keyword_as_source():
    assert_table_lineage_equal(
        "SELECT col1, col2 FROM segment", {"segment"}, test_sqlparse=False
    )


def test_select_in_parenthesis():
    assert_table_lineage_equal("(SELECT * FROM tab1)", {"tab1"}, test_sqlparse=False)



import re

from sqlparse import tokens
from sqlparse.engine import grouping
from sqlparse.keywords import KEYWORDS, SQL_REGEX

from sqllineage.core.parser.sqlparse.utils import group_function_with_window


def _patch_adding_window_function_token() -> None:
    grouping.group_functions = group_function_with_window


def _patch_adding_builtin_type() -> None:
    KEYWORDS["STRING"] = tokens.Name.Builtin
    KEYWORDS["DATETIME"] = tokens.Name.Builtin


def _patch_updating_lateral_view_lexeme() -> None:
    for i, (regex, lexeme) in enumerate(SQL_REGEX):
        rgx = re.compile(regex, re.IGNORECASE | re.UNICODE).match
        if rgx("LATERAL VIEW EXPLODE(col)"):
            new_regex = r"(LATERAL\s+VIEW\s+)(OUTER\s+)?(EXPLODE|INLINE|PARSE_URL_TUPLE|POSEXPLODE|STACK|JSON_TUPLE)\b"
            SQL_REGEX[i] = (new_regex, lexeme)
            break


def _monkey_patch() -> None:
    _patch_adding_window_function_token()
    _patch_adding_builtin_type()
    _patch_updating_lateral_view_lexeme()


_monkey_patch()



import argparse
import logging
import logging.config
import warnings

from sqllineage import (
    DEFAULT_DIALECT,
    DEFAULT_HOST,
    DEFAULT_LOGGING,
    DEFAULT_PORT,
    NAME as MAIN_NAME,
    VERSION as MAIN_VERSION,
)
from sqllineage.core.metadata.dummy import DummyMetaDataProvider
from sqllineage.core.metadata.sqlalchemy import SQLAlchemyMetaDataProvider
from sqllineage.drawing import draw_lineage_graph
from sqllineage.runner import LineageRunner
from sqllineage.utils.constant import LineageLevel
from sqllineage.utils.helpers import extract_sql_from_args

logger = logging.getLogger(__name__)


def main(args=None) -> None:
    """
    The command line interface entry point.

    :param args: the command line arguments for sqllineage command
    """
    logging.config.dictConfig(DEFAULT_LOGGING)

    parser = argparse.ArgumentParser(
        prog="sqllineage", description="SQL Lineage Parser."
    )
    parser.add_argument(
        "--version", action="version", version="%s %s" % (MAIN_NAME, MAIN_VERSION)
    )
    parser.add_argument(
        "-e", metavar="<quoted-query-string>", help="SQL from command line"
    )
    parser.add_argument("-f", metavar="<filename>", help="SQL from files")
    parser.add_argument(
        "-v",
        "--verbose",
        help="increase output verbosity, show statement level lineage result",
        action="store_true",
    )
    parser.add_argument(
        "-l",
        "--level",
        help="lineage level, column or table, default at table level",
        choices=[LineageLevel.TABLE, LineageLevel.COLUMN],
        default=LineageLevel.TABLE,
    )
    parser.add_argument(
        "-g",
        "--graph-visualization",
        help="show graph visualization of the lineage in a webserver",
        action="store_true",
    )
    parser.add_argument(
        "-H",
        "--host",
        help="the host visualization webserver will be bind to",
        type=str,
        default=DEFAULT_HOST,
        metavar="<hostname>",
    )
    parser.add_argument(
        "-p",
        "--port",
        help="the port visualization webserver will be listening on",
        type=int,
        default=DEFAULT_PORT,
        metavar="<port_number>{0..65536}",
    )
    parser.add_argument(
        "-d",
        "--dialect",
        help="the dialect used to analyze the lineage, use --dialects to show all available dialects",
        type=str,
        default=DEFAULT_DIALECT,
        metavar="<dialect>",
    )
    parser.add_argument(
        "-ds",
        "--dialects",
        help="list all the available dialects",
        action="store_true",
    )
    parser.add_argument(
        "--silent_mode",
        help="skip unsupported statements",
        action="store_true",
    )
    parser.add_argument(
        "--sqlalchemy_url",
        help="sqlalchemy url to provide metadata for lineage analysis",
        type=str,
    )
    args = parser.parse_args(args)
    if args.e and args.f:
        warnings.warn("Both -e and -f options are specified. -e option will be ignored")
    if args.f or args.e:
        sql = extract_sql_from_args(args)
        runner = LineageRunner(
            sql,
            dialect=args.dialect,
            metadata_provider=(
                SQLAlchemyMetaDataProvider(args.sqlalchemy_url)
                if args.sqlalchemy_url
                else DummyMetaDataProvider()
            ),
            verbose=args.verbose,
            draw_options={
                "host": args.host,
                "port": args.port,
                "f": args.f if args.f else None,
            },
            silent_mode=args.silent_mode,
        )
        if args.graph_visualization:
            runner.draw()
        elif args.level == LineageLevel.COLUMN:
            runner.print_column_lineage()
        else:
            runner.print_table_lineage()
    elif args.graph_visualization:
        return draw_lineage_graph(**{"host": args.host, "port": args.port})
    elif args.dialects:
        dialects = []
        for _, supported_dialects in LineageRunner.supported_dialects().items():
            dialects += supported_dialects
        print("\n".join(dialects))
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import json
import os.path
from collections import namedtuple
from http import HTTPStatus
from io import StringIO

from sqllineage.config import SQLLineageConfig
from sqllineage.drawing import app


def test_handler():
    container = namedtuple("response", ["status", "header"])

    def start_response(status, header):
        container.status = status
        container.header = header

    def mock_request(method, path, body=None):
        if isinstance(body, dict):
            body = json.dumps(body)
        environ = {"REQUEST_METHOD": method, "PATH_INFO": path}
        if body:
            with StringIO() as f:
                length = f.write(body)
                f.seek(0)
                environ["CONTENT_LENGTH"] = length
                environ["wsgi.input"] = f
                app(environ, start_response)
        else:
            app(environ, start_response)

    # 200
    mock_request("GET", "/")
    assert container.status.startswith(str(HTTPStatus.OK.value))
    mock_request("GET", "/manifest.json")
    assert container.status.startswith(str(HTTPStatus.OK.value))
    mock_request("POST", "/lineage", {"e": "SELECT * FROM dual", "p": 5000})
    assert container.status.startswith(str(HTTPStatus.OK.value))
    mock_request("POST", "/script", {"e": "SELECT * FROM dual", "p": 5000})
    assert container.status.startswith(str(HTTPStatus.OK.value))
    mock_request(
        "POST",
        "/directory",
        {"f": os.path.join(SQLLineageConfig.DIRECTORY, "tpcds/query01.sql")},
    )
    assert container.status.startswith(str(HTTPStatus.OK.value))
    mock_request(
        "POST",
        "/directory",
        {"d": os.path.join(SQLLineageConfig.DIRECTORY, "tpcds/")},
    )
    assert container.status.startswith(str(HTTPStatus.OK.value))
    mock_request("POST", "/directory", {})
    assert container.status.startswith(str(HTTPStatus.OK.value))
    mock_request("OPTIONS", "/directory", {})
    assert container.status.startswith(str(HTTPStatus.OK.value))
    # 400
    mock_request("POST", "/lineage", {"e": "SELECT * FROM where foo='bar'"})
    assert container.status.startswith(str(HTTPStatus.BAD_REQUEST.value))
    # 403
    mock_request("POST", "/directory", {"f": "/etc/passwd"})
    assert container.status.startswith(str(HTTPStatus.FORBIDDEN.value))
    mock_request("POST", "/directory", {"d": "/"})
    assert container.status.startswith(str(HTTPStatus.FORBIDDEN.value))
    # 404
    mock_request("GET", "/non-exist-resource")
    assert container.status.startswith(str(HTTPStatus.NOT_FOUND.value))
    mock_request("GET", "/static")
    assert container.status.startswith(str(HTTPStatus.NOT_FOUND.value))
    mock_request(
        "POST",
        "/script",
        {"f": os.path.join(SQLLineageConfig.DIRECTORY, "tpcds/query100.sql")},
    )
    assert container.status.startswith(str(HTTPStatus.NOT_FOUND.value))
    mock_request("POST", "/non-exist-resource", {"e": "SELECT * FROM where foo='bar'"})
    assert container.status.startswith(str(HTTPStatus.NOT_FOUND.value))
    mock_request("GET", "/../cli.py")
    assert container.status.startswith(str(HTTPStatus.NOT_FOUND.value))
    mock_request("OPTIONS", "/non-exist-resource")
    assert container.status.startswith(str(HTTPStatus.NOT_FOUND.value))
    # 405
    mock_request("PUT", "/")
    assert container.status.startswith(str(HTTPStatus.METHOD_NOT_ALLOWED.value))
    mock_request("PATCH", "/")
    assert container.status.startswith(str(HTTPStatus.METHOD_NOT_ALLOWED.value))
    mock_request("DELETE", "/")
    assert container.status.startswith(str(HTTPStatus.METHOD_NOT_ALLOWED.value))



"""
specify columns in CREATE/INSERT statement.
DO NOT support this feature with non-validating dialect because sqlparse would parse table/view name as function call
"""

from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal, assert_table_lineage_equal


def test_view_with_subquery_custom_columns():
    # select as subquery
    sql = "CREATE VIEW my_view (random1,random2) AS (SELECT col1,col2 FROM tbl)"
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tbl"),
                ColumnQualifierTuple("random1", "my_view"),
            ),
            (
                ColumnQualifierTuple("col2", "tbl"),
                ColumnQualifierTuple("random2", "my_view"),
            ),
        ],
        test_sqlparse=False,
    )
    sql = "CREATE VIEW my_view (random1,random2) AS ((SELECT col1,col2 FROM tbl))"
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tbl"),
                ColumnQualifierTuple("random1", "my_view"),
            ),
            (
                ColumnQualifierTuple("col2", "tbl"),
                ColumnQualifierTuple("random2", "my_view"),
            ),
        ],
        test_sqlparse=False,
    )
    sql = "CREATE VIEW my_view (random1,random2) AS (((SELECT col1,col2 FROM tbl)))"
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tbl"),
                ColumnQualifierTuple("random1", "my_view"),
            ),
            (
                ColumnQualifierTuple("col2", "tbl"),
                ColumnQualifierTuple("random2", "my_view"),
            ),
        ],
        test_sqlparse=False,
    )
    sql = "CREATE VIEW my_view (random1,random2) AS SELECT col1,col2 FROM tbl"
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tbl"),
                ColumnQualifierTuple("random1", "my_view"),
            ),
            (
                ColumnQualifierTuple("col2", "tbl"),
                ColumnQualifierTuple("random2", "my_view"),
            ),
        ],
        test_sqlparse=False,
    )


def test_create_view_with_same_columns():
    sql = "CREATE VIEW my_view (col1,col2) AS (SELECT col1,col2 FROM tbl)"
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tbl"),
                ColumnQualifierTuple("col1", "my_view"),
            ),
            (
                ColumnQualifierTuple("col2", "tbl"),
                ColumnQualifierTuple("col2", "my_view"),
            ),
        ],
        test_sqlparse=False,
    )


def test_insert_into_with_columns():
    # table lineage works for sqlparse
    assert_table_lineage_equal(
        "INSERT INTO tab1 (col1, col2) SELECT * FROM tab2;",
        {"tab2"},
        {"tab1"},
        test_sqlparse=False,
    )


def test_insert_into_with_columns_and_select_union():
    assert_table_lineage_equal(
        "INSERT INTO tab1 (col1, col2) SELECT * FROM tab2 UNION SELECT * FROM tab3",
        {"tab2", "tab3"},
        {"tab1"},
        test_sqlparse=False,
    )
    assert_table_lineage_equal(
        "INSERT INTO tab1 (col1, col2) (SELECT * FROM tab2 UNION SELECT * FROM tab3)",
        {"tab2", "tab3"},
        {"tab1"},
        test_sqlparse=False,
    )
    assert_table_lineage_equal(
        "INSERT INTO tab1 (col1, col2) ((SELECT * FROM tab2 UNION SELECT * FROM tab3))",
        {"tab2", "tab3"},
        {"tab1"},
        test_sqlparse=False,
    )
    assert_table_lineage_equal(
        "INSERT INTO tab1 (col1, col2) (((SELECT * FROM tab2 UNION SELECT * FROM tab3)))",
        {"tab2", "tab3"},
        {"tab1"},
        test_sqlparse=False,
    )


def test_insert_with_custom_columns():
    # test with query as subquery
    sql = "INSERT INTO tgt_tbl(random1, random2) (SELECT col1,col2 FROM src_tbl)"
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "src_tbl"),
                ColumnQualifierTuple("random1", "tgt_tbl"),
            ),
            (
                ColumnQualifierTuple("col2", "src_tbl"),
                ColumnQualifierTuple("random2", "tgt_tbl"),
            ),
        ],
        test_sqlparse=False,
    )

    # test with plain query
    sql = "INSERT INTO tgt_tbl(random1, random2) SELECT col1,col2 FROM src_tbl_new"
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "src_tbl_new"),
                ColumnQualifierTuple("random1", "tgt_tbl"),
            ),
            (
                ColumnQualifierTuple("col2", "src_tbl_new"),
                ColumnQualifierTuple("random2", "tgt_tbl"),
            ),
        ],
        test_sqlparse=False,
    )


def test_insert_with_custom_columns_and_cte_within_query():
    sql = """INSERT INTO tab2 (col1)
WITH cte1 AS (SELECT col2 FROM tab1)
SELECT col2 FROM cte1"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col2", "tab1"),
                ColumnQualifierTuple("col1", "tab2"),
            )
        ],
        test_sqlparse=False,
    )



import pytest

from ....helpers import assert_table_lineage_equal


@pytest.mark.parametrize("dialect", ["tsql"])
def test_tsql_declare(dialect: str):
    """
    https://learn.microsoft.com/en-us/sql/t-sql/language-elements/declare-local-variable-transact-sql?view=sql-server-ver16
    """
    sql = """DECLARE @age int = 10;
INSERT INTO tgt
SELECT Name,@age
FROM People;"""
    assert_table_lineage_equal(sql, {"People"}, {"tgt"}, dialect=dialect)



import os
import threading
from typing import Any, Dict, Set

from sqllineage.exceptions import ConfigException


class _SQLLineageConfigLoader:
    """
    Load all configurable items from environment variable, otherwise fallback to default
    """

    # inspired by https://github.com/joke2k/django-environ
    config = {
        # for frontend directory drawer
        "DIRECTORY": (str, os.path.join(os.path.dirname(__file__), "data")),
        # set default schema/database
        "DEFAULT_SCHEMA": (str, ""),
        # to enable tsql no semicolon splitter mode
        "TSQL_NO_SEMICOLON": (bool, False),
        # lateral column alias reference supported by some dialect (redshift, spark 3.4+, etc)
        "LATERAL_COLUMN_ALIAS_REFERENCE": (bool, False),
    }

    def __init__(self) -> None:
        self._thread_config: Dict[int, Dict[str, Any]] = {}
        self._thread_in_context_manager: Set[int] = set()

    def __getattr__(self, item: str):
        if item in self.config.keys():
            if (
                value := self._thread_config.get(self.get_ident(), {}).get(item)
            ) is not None:
                return value

            type_, default = self.config[item]
            # require SQLLINEAGE_ prefix from environment variable
            return self.parse_value(
                os.environ.get("SQLLINEAGE_" + item, default), type_
            )
        else:
            return super().__getattribute__(item)

    def __setattr__(self, key, value) -> None:
        if key in self.config:
            raise ConfigException(
                "SQLLineageConfig is read-only. Use context manager to update thread level config."
            )
        else:
            super().__setattr__(key, value)

    def __call__(self, *args, **kwargs):
        if self.get_ident() not in self._thread_config.keys():
            self._thread_config[self.get_ident()] = {}
        for key, value in kwargs.items():
            if key in self.config.keys():
                self._thread_config[self.get_ident()][key] = self.parse_value(
                    value, self.config[key][0]
                )
            else:
                raise ConfigException(f"Invalid config key: {key}")
        return self

    def __enter__(self):
        if (thread_id := self.get_ident()) not in self._thread_in_context_manager:
            self._thread_in_context_manager.add(thread_id)
        else:
            raise ConfigException("SQLLineageConfig context manager is not reentrant")

    def __exit__(self, exc_type, exc_val, exc_tb):
        thread_id = self.get_ident()
        if thread_id in self._thread_config:
            self._thread_config.pop(self.get_ident())
        if thread_id in self._thread_in_context_manager:
            self._thread_in_context_manager.remove(thread_id)

    @staticmethod
    def get_ident() -> int:
        return threading.get_ident()

    @staticmethod
    def parse_value(value, cast) -> Any:
        """Parse and cast provided value
        :param value: Stringed value.
        :param cast: Type to cast return value as.
        :returns: cast value
        """
        if cast is bool:
            try:
                value = int(value) != 0
            except ValueError:
                value = value.lower().strip() in ("true", "on", "ok", "y", "yes", "1")
        else:
            value = cast(value)
        return value


SQLLineageConfig = _SQLLineageConfigLoader()



import concurrent.futures
import os
import random
import time
from unittest.mock import patch

import pytest

from sqllineage.config import SQLLineageConfig
from sqllineage.exceptions import ConfigException
from sqllineage.runner import LineageRunner


@patch.dict(
    os.environ,
    {
        "SQLLINEAGE_DIRECTORY": os.path.join(os.path.dirname(__file__), "data"),
        "SQLLINEAGE_DEFAULT_SCHEMA": "<default>",
        "SQLLINEAGE_TSQL_NO_SEMICOLON": "true",
    },
)
def test_config():
    assert type(SQLLineageConfig.DIRECTORY) is str
    assert SQLLineageConfig.DIRECTORY == os.path.join(os.path.dirname(__file__), "data")

    assert type(SQLLineageConfig.DEFAULT_SCHEMA) is str
    assert SQLLineageConfig.DEFAULT_SCHEMA == "<default>"

    assert type(SQLLineageConfig.TSQL_NO_SEMICOLON) is bool
    assert SQLLineageConfig.TSQL_NO_SEMICOLON is True


def test_disable_direct_update_config():
    with pytest.raises(ConfigException):
        SQLLineageConfig.DEFAULT_SCHEMA = "ods"


def test_update_config_using_context_manager():
    with SQLLineageConfig(LATERAL_COLUMN_ALIAS_REFERENCE=True):
        assert SQLLineageConfig.LATERAL_COLUMN_ALIAS_REFERENCE is True
    assert SQLLineageConfig.LATERAL_COLUMN_ALIAS_REFERENCE is False

    with SQLLineageConfig(DEFAULT_SCHEMA="ods"):
        assert SQLLineageConfig.DEFAULT_SCHEMA == "ods"
    assert SQLLineageConfig.DEFAULT_SCHEMA == ""

    with SQLLineageConfig(DIRECTORY=""):
        assert SQLLineageConfig.DIRECTORY == ""
    assert SQLLineageConfig.DIRECTORY != ""


def test_update_config_context_manager_non_reentrant():
    with pytest.raises(ConfigException):
        with SQLLineageConfig(DEFAULT_SCHEMA="ods"):
            with SQLLineageConfig(DEFAULT_SCHEMA="dwd"):
                pass


def test_disable_update_unknown_config():
    with pytest.raises(ConfigException):
        with SQLLineageConfig(UNKNOWN_KEY="value"):
            pass


def _check_schema(schema: str):
    # used by test_config_parallel, must be a global function so that it can be pickled between processes
    with SQLLineageConfig(DEFAULT_SCHEMA=schema):
        table = LineageRunner("select * from test").source_tables.pop()
        # randomly sleep [0, 0.1) second to simulate real parsing scenario
        time.sleep(random.random() * 0.1)
        return table.schema.raw_name


@pytest.mark.parametrize("pool", ["ThreadPoolExecutor", "ProcessPoolExecutor"])
def test_config_parallel(pool: str):
    executor_class = getattr(concurrent.futures, pool)
    schemas = [f"db{i}" for i in range(100)]
    with executor_class() as executor:
        futures = [executor.submit(_check_schema, schema) for schema in schemas]
        for i, future in enumerate(futures):
            assert future.result() == schemas[i]



from typing import Dict, List, Optional

from sqllineage.core.metadata_provider import MetaDataProvider


class DummyMetaDataProvider(MetaDataProvider):
    """
    A Dummy MetaDataProvider that accept metadata as a dict
    """

    def __init__(self, metadata: Optional[Dict[str, List[str]]] = None):
        """
        :param metadata: a dict with schema.table name as key and a list of unqualified column name as value
        """
        super().__init__()
        self.metadata = metadata if metadata is not None else {}

    def _get_table_columns(self, schema: str, table: str, **kwargs) -> List[str]:
        return self.metadata.get(f"{schema}.{table}", [])

    def __bool__(self):
        return len(self.metadata) > 0



from typing import Optional, Union

from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import StatementLineageHolder
from sqllineage.core.models import Column, SubQuery, Table
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.core.parser.sqlfluff.extractors.cte import CteExtractor
from sqllineage.core.parser.sqlfluff.extractors.select import SelectExtractor
from sqllineage.core.parser.sqlfluff.models import SqlFluffSubQuery
from sqllineage.core.parser.sqlfluff.utils import (
    extract_column_qualifier,
    extract_identifier,
    extract_innermost_bracketed,
    list_child_segments,
)
from sqllineage.utils.entities import AnalyzerContext


class MergeExtractor(BaseExtractor):
    """
    Merge statement lineage extractor
    """

    SUPPORTED_STMT_TYPES = ["merge_statement"]

    def extract(
        self,
        statement: BaseSegment,
        context: AnalyzerContext,
    ) -> StatementLineageHolder:
        holder = StatementLineageHolder()
        src_flag = tgt_flag = False
        direct_source: Optional[Union[Table, SubQuery]] = None
        segments = list_child_segments(statement)
        for i, segment in enumerate(segments):
            if segment.type == "merge_match":
                merge_match = segment
                for merge_when_matched_clause in merge_match.get_children(
                    "merge_when_matched_clause"
                ):
                    if merge_update_clause := merge_when_matched_clause.get_child(
                        "merge_update_clause"
                    ):
                        if set_clause_list := merge_update_clause.get_child(
                            "set_clause_list"
                        ):
                            for set_clause in set_clause_list.get_children(
                                "set_clause"
                            ):
                                columns = set_clause.get_children("column_reference")
                                if len(columns) == 2:
                                    src_col = tgt_col = None
                                    if src_cqt := extract_column_qualifier(columns[1]):
                                        src_col = Column(src_cqt.column)
                                        src_col.parent = direct_source
                                    if tgt_cqt := extract_column_qualifier(columns[0]):
                                        tgt_col = Column(tgt_cqt.column)
                                        tgt_col.parent = list(holder.write)[0]
                                    if src_col is not None and tgt_col is not None:
                                        holder.add_column_lineage(src_col, tgt_col)
                for merge_when_not_matched_clause in merge_match.get_children(
                    "merge_when_not_matched_clause"
                ):
                    if merge_insert := merge_when_not_matched_clause.get_child(
                        "merge_insert_clause"
                    ):
                        insert_columns = []
                        if bracketed := merge_insert.get_child("bracketed"):
                            for column_reference in bracketed.get_children(
                                "column_reference"
                            ):
                                if cqt := extract_column_qualifier(column_reference):
                                    tgt_col = Column(cqt.column)
                                    tgt_col.parent = list(holder.write)[0]
                                    insert_columns.append(tgt_col)
                            if values_clause := merge_insert.get_child("values_clause"):
                                if bracketed := values_clause.get_child("bracketed"):
                                    for j, e in enumerate(
                                        bracketed.get_children("literal", "expression")
                                    ):
                                        if column_reference_optional := e.get_child(
                                            "column_reference"
                                        ):
                                            if cqt := extract_column_qualifier(
                                                column_reference_optional
                                            ):
                                                src_col = Column(cqt.column)
                                                src_col.parent = direct_source
                                                holder.add_column_lineage(
                                                    src_col, insert_columns[j]
                                                )
            elif segment.type == "keyword":
                if segment.raw_upper in ["MERGE", "INTO"]:
                    tgt_flag = True
                elif segment.raw_upper == "USING":
                    src_flag = True
                continue

            if tgt_flag:
                if table := self.find_table(segment):
                    holder.add_write(table)
                tgt_flag = False
            if src_flag:
                if table := self.find_table(segment):
                    holder.add_read(table)
                    direct_source = table
                elif segment.type == "bracketed":
                    next_segment = segments[i + 1]
                    direct_source = SqlFluffSubQuery.of(
                        extract_innermost_bracketed(segment),
                        (
                            extract_identifier(next_segment)
                            if next_segment.type == "alias_expression"
                            else None
                        ),
                    )
                    holder.add_read(direct_source)

                    extractor_cls = (
                        CteExtractor
                        if direct_source.query.get_child(
                            "with_compound_statement"
                        )  # in case the subquery is a CTE query
                        else SelectExtractor  # in case the subquery is a select query
                    )
                    holder |= self.delegate_to(
                        extractor_cls,
                        direct_source.query,
                        AnalyzerContext(cte=holder.cte, write={direct_source}),
                    )
                src_flag = False
        return holder



import warnings
from typing import Dict, List

from sqlfluff.core import (
    FluffConfig,
    Linter,
    SQLLexError,
    SQLParseError,
    dialect_readout,
)
from sqlfluff.core.parser import BaseSegment

from sqllineage.core.analyzer import LineageAnalyzer
from sqllineage.core.holders import StatementLineageHolder
from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.exceptions import (
    InvalidSyntaxException,
    UnsupportedStatementException,
)
from sqllineage.utils.entities import AnalyzerContext


class SqlFluffLineageAnalyzer(LineageAnalyzer):
    """SQL Statement Level Lineage Analyzer for `sqlfluff`"""

    PARSER_NAME = "sqlfluff"
    SUPPORTED_DIALECTS = list(dialect.label for dialect in dialect_readout())

    def __init__(self, dialect: str, silent_mode: bool = False):
        self._dialect = dialect
        self._silent_mode = silent_mode
        self.tsql_split_cache: Dict[str, BaseSegment] = {}

    def split_tsql(self, sql: str) -> List[str]:
        """
        use sqlfluff parse to split tsql statements. This is in particular for semicolon not present cases.
        The result is cached so that later analyze method doesn't have to parse regarding single statement sql.
        """
        sqls = []
        for segment in self._list_specific_statement_segment(sql):
            self.tsql_split_cache[segment.raw] = segment
            sqls.append(segment.raw)
        return sqls

    def analyze(
        self, sql: str, metadata_provider: MetaDataProvider
    ) -> StatementLineageHolder:
        if sql in self.tsql_split_cache:
            statement_segments = [self.tsql_split_cache[sql]]
        else:
            statement_segments = self._list_specific_statement_segment(sql)
        if len(statement_segments) == 0:
            raise UnsupportedStatementException(
                f"SQLLineage cannot parse SQL:" f"{sql}"
            )  # pragma: no cover
        else:
            statement_segment = statement_segments[0]
            for extractor in [
                extractor_cls(self._dialect, metadata_provider)
                for extractor_cls in BaseExtractor.__subclasses__()
            ]:
                if extractor.can_extract(statement_segment.type):
                    lineage_holder = extractor.extract(
                        statement_segment, AnalyzerContext()
                    )
                    return StatementLineageHolder.of(lineage_holder)
            else:
                if self._silent_mode:
                    warnings.warn(
                        f"SQLLineage doesn't support analyzing statement type [{statement_segment.type}] for SQL:"
                        f"{sql}"
                    )
                    return StatementLineageHolder()
                else:
                    raise UnsupportedStatementException(
                        f"SQLLineage doesn't support analyzing statement type [{statement_segment.type}] for SQL:"
                        f"{sql}"
                    )

    def _list_specific_statement_segment(self, sql: str):
        parsed = Linter(
            config=FluffConfig.from_root(overrides={"dialect": self._dialect})
        ).parse_string(sql)
        violations = [
            str(e)
            for e in parsed.violations
            if isinstance(e, (SQLLexError, SQLParseError))
        ]
        if violations:
            violation_msg = "\n".join(violations)
            raise InvalidSyntaxException(
                f"This SQL statement is unparsable, please check potential syntax error for SQL:\n"
                f"{sql}\n"
                f"{violation_msg}"
            )
        segments = []
        for top_segment in getattr(parsed.tree, "segments", []):
            if top_segment.type == "statement":
                segments.append(top_segment.segments[0])
            elif top_segment.type == "batch":
                statements = top_segment.get_children("statement")
                if len(statements) > 1:
                    warnings.warn(
                        "SQL statements is not split by semicolon. "
                        "SQLLineage is not guaranteed to generate correct result under this circumstances.",
                        SyntaxWarning,
                        stacklevel=2,
                    )
                for statement in statements:
                    segments.append(statement.segments[0])
        return segments



class NodeTag:
    READ = "read"
    WRITE = "write"
    CTE = "cte"
    DROP = "drop"
    SOURCE_ONLY = "source_only"
    TARGET_ONLY = "target_only"
    SELFLOOP = "selfloop"


class EdgeTag:
    INDEX = "index"


class EdgeType:
    LINEAGE = "lineage"
    RENAME = "rename"
    HAS_COLUMN = "has_column"
    HAS_ALIAS = "has_alias"


class LineageLevel:
    TABLE = "table"
    COLUMN = "column"



from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.models import Path
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.core.parser.sqlfluff.extractors.select import SelectExtractor
from sqllineage.core.parser.sqlfluff.models import SqlFluffColumn, SqlFluffTable
from sqllineage.core.parser.sqlfluff.utils import (
    list_child_segments,
)
from sqllineage.utils.entities import AnalyzerContext
from sqllineage.utils.helpers import escape_identifier_name


class CreateInsertExtractor(BaseExtractor):
    """
    Create statement and Insert statement lineage extractor
    """

    SUPPORTED_STMT_TYPES = [
        "create_table_statement",
        "create_table_as_statement",
        "create_view_statement",
        "insert_statement",
        "insert_overwrite_directory_hive_fmt_statement",
    ]

    def extract(
        self,
        statement: BaseSegment,
        context: AnalyzerContext,
    ) -> SubQueryLineageHolder:
        holder = self._init_holder(context)
        src_flag = tgt_flag = False
        for segment in list_child_segments(statement):
            if segment.type == "with_compound_statement":
                holder |= self.delegate_to_cte(segment, holder)
            elif segment.type == "bracketed" and any(
                s.type == "with_compound_statement" for s in segment.segments
            ):
                for sgmt in segment.segments:
                    if sgmt.type == "with_compound_statement":
                        holder |= self.delegate_to_cte(segment, holder)
            elif segment.type in ("select_statement", "set_expression"):
                holder |= self.delegate_to_select(segment, holder)
            elif segment.type == "values_clause":
                for bracketed in segment.get_children("bracketed"):
                    for expression in bracketed.get_children("expression"):
                        if sub_bracketed := expression.get_child("bracketed"):
                            if sub_expression := sub_bracketed.get_child("expression"):
                                if select_statement := sub_expression.get_child(
                                    "select_statement"
                                ):
                                    holder |= self.delegate_to_select(
                                        select_statement, holder
                                    )
            elif segment.type == "bracketed" and (
                subquery_segments := list(
                    segment.recursive_crawl(
                        "select_statement", "set_expression", recurse_into=False
                    )
                )
            ):
                # note regular subquery within SELECT statement is handled by SelectExtractor, this is only to handle
                # top-level subquery in DML like: 1) create table foo as (subquery); 2) insert into foo (subquery)
                # subquery here isn't added as read source, and it inherits DML-level write_columns if parsed
                for subquery_segment in subquery_segments:
                    holder |= self.delegate_to_select(subquery_segment, holder)

            elif segment.type == "bracketed":
                # In case of bracketed column reference, add these target columns to holder
                # so that when we compute the column level lineage
                # we keep these columns into consideration
                sub_segments = list_child_segments(segment)
                if all(
                    sub_segment.type in ["column_reference", "column_definition"]
                    for sub_segment in sub_segments
                ):
                    # target columns only apply to bracketed column_reference and column_definition
                    columns = []
                    for sub_segment in sub_segments:
                        if sub_segment.type == "column_definition":
                            if identifier := sub_segment.get_child("identifier"):
                                sub_segment = identifier
                        columns.append(SqlFluffColumn.of(sub_segment))
                    holder.add_write_column(*columns)

            elif segment.type == "keyword":
                if segment.raw_upper in [
                    "INTO",
                    "OVERWRITE",
                    "TABLE",
                    "VIEW",
                    "DIRECTORY",
                ] or (
                    tgt_flag is True and segment.raw_upper in ["IF", "NOT", "EXISTS"]
                ):
                    tgt_flag = True
                elif segment.raw_upper in ["LIKE", "CLONE"]:
                    src_flag = True
                continue

            if tgt_flag:
                if segment.type in ["table_reference", "object_reference"]:
                    write_obj = SqlFluffTable.of(segment)
                    holder.add_write(write_obj)
                elif segment.type == "literal":
                    if segment.raw.isnumeric():
                        # Special Handling for Spark Bucket Table DDL
                        pass
                    else:
                        holder.add_write(Path(escape_identifier_name(segment.raw)))
                tgt_flag = False
            if src_flag:
                if segment.type in ["table_reference", "object_reference"]:
                    holder.add_read(SqlFluffTable.of(segment))
                src_flag = False
        return holder

    def delegate_to_cte(
        self, segment: BaseSegment, holder: SubQueryLineageHolder
    ) -> SubQueryLineageHolder:
        from .cte import CteExtractor

        return self.delegate_to(
            CteExtractor,
            segment,
            AnalyzerContext(
                cte=holder.cte, write=holder.write, write_columns=holder.write_columns
            ),
        )

    def delegate_to_select(
        self,
        segment: BaseSegment,
        holder: SubQueryLineageHolder,
    ) -> SubQueryLineageHolder:
        return self.delegate_to(
            SelectExtractor,
            segment,
            AnalyzerContext(
                cte=holder.cte,
                write=holder.write,
                write_columns=holder.write_columns,
            ),
        )
